/*Copyright and Author
 *
 *This program is free software, distributed under the terms of the GNU General Public License Version 2.
 *
 * \author Leena <leena_m@spanservices.com>, (SPAN Infotech India Pvt. Ltd., Bangalore)
 *
 * \module diameter
 * 
 * \created 24\03\09
 */

/*! \file
  \brief Handling avp related routines.
  
 */

#include "asterisk.h"
#include "asterisk/cdr.h"
#include "diameter_base.h"

static void set_avpflag_and_type(struct avp *avp);

/*! \brief Creates an AVP adds it into a list.*/
int build_and_add_avp(int avp_code, const char* avp_data, unsigned int avpdata_length, struct avp** avp_list, unsigned int vendor_id, enum avp_flag avpflag_value)
{
	struct avp *avp_t = NULL;
	struct avp *pavp  = *avp_list;
	
	if (!(avp_list && avp_data))
		return 0;

	if (!(avp_t = (struct avp*)malloc(sizeof(struct avp)))) 
    return 0;
    
	if (!(avp_t->data = (char*)malloc(avpdata_length + 1))) {
		free(avp_t);
		avp_t = NULL;
		return 0;
	}

	avp_t->code = avp_code;
	avp_t->flag = avpflag_value;
	avp_t->vendor_id = vendor_id;
	set_avpflag_and_type(avp_t);
	strncpy(avp_t->data, avp_data, avpdata_length);
	avp_t->data[avpdata_length] = '\0';
	avp_t->length = avpdata_length;
	avp_t->next = NULL;

	/* inserting the newly created avp as the first node of the list */
	if (pavp) 
		avp_t->next = pavp;
	pavp = avp_t;

	*avp_list = pavp;
	return 1;
}
	
/*!
* \brief Sets the flag and type of an AVP.
*	\note The flag is set only if it is not set.If the flag of the avp already has a value it is untouched(when a received
* diameter message is parsed then the avp flag will be set already when calling this function.
*
* \param avp the avp whose flag and type is set
*
*/
static void set_avpflag_and_type(struct avp *avp)
{
	switch (avp->code) {
	case AVP_USER_NAME: 	
  case AVP_SESSION_ID: 
  case AVP_DESTINATION_REALM: 
  case AVP_DESTINATION_HOST: 
  case AVP_ORIGIN_HOST: 
	case AVP_ORIGIN_REALM: 
		if (!avp->flag)
			avp->flag = AVP_FLAG_MANDATORY | avp->flag; /* The M bit is set indicating the AVP is mandatory.*/
		avp->type = AVP_STRING_TYPE;
		break;

	case AVP_VENDOR_ID:
	case AVP_RESULT_CODE: 
		if (!avp->flag)
			avp->flag = AVP_FLAG_MANDATORY | avp->flag;
		avp->type = AVP_INTEGER32_TYPE;
		break;

	case AVP_TERMINATION_CAUSE:
	case AVP_DISCONNECT_CAUSE:
	case AVP_ACCT_RECORD_TYPE: 
		if (!avp->flag)
			avp->flag = AVP_FLAG_MANDATORY | avp->flag;
		avp->type = AVP_ENUMERATED_TYPE;
		break;

	case AVP_AUTH_APPLICATION_ID:	
	case AVP_ACCT_APPLICATION_ID:
	case AVP_ACCT_RECORD_NUMBER:
		if (!avp->flag)
			avp->flag = AVP_FLAG_MANDATORY | avp->flag;
		avp->type = AVP_UNSIGNED32_TYPE;
		break;

	case AVP_HOST_IP_ADDRESS: 
		if (!avp->flag)
			avp->flag = AVP_FLAG_MANDATORY | avp->flag;
		avp->type = AVP_ADDRESS_TYPE;
    break;

	case AVP_PRODUCT_NAME:
		if (!avp->flag)
			avp->flag = AVP_FLAG_MANDATORY | avp->flag;
		avp->type = AVP_STRING_TYPE;
		break;

	case AVP_ACCOUNTING_SUB_SESSION_ID:
		if (!avp->flag)
			avp->flag = AVP_FLAG_MANDATORY | avp->flag;
		avp->type = AVP_UNSIGNED64_TYPE;
    break;

	case AVP_FAILED:
		if (!avp->flag)
			avp->flag = AVP_FLAG_NONE;
		avp->type = AVP_GROUPED_TYPE;
		break;

	/* Asterisk specific AVPs */
	case AVP_AST_ACCT_CODE:
	case AVP_AST_SRC:
	case AVP_AST_DST:
	case AVP_AST_DST_CTX:
	case AVP_AST_CLID:
	case AVP_AST_CHAN:
	case AVP_AST_DST_CHAN:
	case AVP_AST_LAST_APP:
	case AVP_AST_LAST_DATA:
	case AVP_AST_UNIQUE_ID:
	case AVP_AST_USER_FIELD:
		if (!avp->flag)
			avp->flag = AVP_FLAG_NONE;
		avp->type = AVP_STRING_TYPE;
		break;

	case AVP_AST_START_TIME:
	case AVP_AST_ANSWER_TIME:
	case AVP_AST_END_TIME:
	case AVP_AST_DURATION:
	case AVP_AST_BILL_SEC:
	case AVP_AST_DISPOSITION:
	case AVP_AST_AMA_FLAGS:
		if (!avp->flag)
			avp->flag = AVP_FLAG_NONE;
		avp->type = AVP_INTEGER32_TYPE;
		break;
		
	/* If the requested AVP is not handled the following is set as flag and type to it. */	
	default:
		ast_debug(1, "Type and Flag for the Code is not defined.\n");
			if (avp->flag && AVP_FLAG_MANDATORY) {
				avp->flag = AVP_FLAG_UNDEFINED;
				avp->type = AVP_TYPE_UNDEFINED;
			}
			break;
	}
}

/*! \brief Builds and adds the specific mandatory avps of a diameter message.*/
int build_mandatory_avps(int msg_code, struct avp** avplist)
{
	char* avp_val = NULL;
	char buf[100] = "\0";

	if (!avplist)
		return 0;

	if (MSG_ACR == msg_code) {
		avp_val = diameter_base->config->peer_realm;
		if	(!build_and_add_avp(AVP_DESTINATION_REALM, avp_val, strlen(avp_val), avplist, VENDOR_CODE, 0))
			return 0;	
		
		sprintf(buf, "%d", 0);
		if (!build_and_add_avp(AVP_ACCOUNTING_SUB_SESSION_ID, buf, strlen(buf), avplist, VENDOR_CODE, 0))
			return 0;
	
		sprintf(buf, "%d", DIAMETER_BASE_ACCOUNTING);
		if (!build_and_add_avp(AVP_ACCT_APPLICATION_ID, buf, strlen(buf), avplist, VENDOR_CODE, 0))
			return 0;

	}	else if	(MSG_CER == msg_code)	{
		avp_val = diameter_base->config->host_ip;
		if (!build_and_add_avp(AVP_HOST_IP_ADDRESS, avp_val, strlen(avp_val), avplist, VENDOR_CODE, 0))
			return 0;

		sprintf(buf, "%d", VENDOR_CODE);
		if (!build_and_add_avp(AVP_VENDOR_ID, buf, strlen(buf), avplist, VENDOR_CODE, 0))
			return 0;

		avp_val = diameter_base->config->host_productname;
		if (!build_and_add_avp(AVP_PRODUCT_NAME, avp_val, strlen(avp_val), avplist, VENDOR_CODE, 0))
			return 0;

		sprintf(buf, "%d", DIAMETER_BASE_ACCOUNTING);
		if (!build_and_add_avp(AVP_ACCT_APPLICATION_ID, buf, strlen(buf), avplist, VENDOR_CODE, 0))
			return 0;

	}	else {
		ast_debug(5, "The requested message is not handled.\n");
	}	
	return 1;
}	

/*! \brief Looks for the presence of Origin_Host and Origin_Realm Avps in an avp list.*/
int check_host_nrealm_presence(struct avp** avp_list)
{
	struct avp* avp = *avp_list;

	if (check_avp_presence(&avp, AVP_ORIGIN_HOST)) {
		if (check_avp_presence(&avp, AVP_ORIGIN_REALM))
			return 1;
	}
	return 0;
}		

/*! \brief Looks for an avp in an avp list. */
int check_avp_presence(struct avp** avp_list, enum avp_code_num avpcode)
{
	struct avp *avp_t = NULL;

	avp_t = *avp_list;
	while	(NULL != avp_t)	{
		if (avpcode == avp_t->code)
			return 1;
		avp_t = avp_t->next;
	}
	return 0;
}

/*! \brief Returns 1 if the ACCT_REALTIME_REQUIRED avp's value is same as parameter avp_val else 0. */
int realtime_avpcheck(struct avp* avp_p , enum avp_acct_realtime avp_val)
{
	char* avpval = NULL;

	if (NULL != (avpval = get_avpval(&(avp_p), AVP_ACCT_REALTIME_REQUIRED))) {
		if (avp_val == atoi(avpval))
			return 1;
	}
	return 0;
}

/* !\brief Returns the data of an avp if present in an avp list else 0. */
char* get_avpval(struct avp** avplist, enum avp_code_num avpcode)
{
	struct avp* avp_t = *avplist;

	while	(avp_t != NULL)	{
		if (avpcode == avp_t->code)	{
			return avp_t->data;
		}
		avp_t = avp_t->next;
	}
	return NULL;
}

/* !\brief Cleans up an avplist. */
void avplist_cleanup(struct avp* avp_list)
{
	struct avp* avp_p = NULL;

	while	(NULL != avp_list) {
  	avp_p = avp_list;
    avp_list = avp_list->next;

   	if (avp_p->data) {
    	free(avp_p->data);
	    avp_p->data = NULL;
		}
		free(avp_p);
		avp_p = NULL;
	}
}	
