/*Copyright and Author
 *
 *This program is free software, distributed under the terms of the GNU General Public License Version 2.
 *
 * \author Leena <leena_m@spanservices.com>, (SPAN Infotech India Pvt. Ltd., Bangalore)
 *
 * \module diameter
 * 
 * \created 06\03\09
 */

/*! \file
  \brief Core diameter client routines.
  
 */

#include "asterisk.h"
#include "asterisk/utils.h"
#include "asterisk/hashtab.h"
#include "asterisk/sched.h"
#include "diameter_base.h"
#include "diameter_transaction.h"

/*! \brief The max. no. of accounting sessions at any point of time */
#define MAX_CALLS 20

/*! \global variable for diameter client structure.*/
struct diameter_base *diameter_base;

/*! \function declarations */
static int get_fqdn(char** fqdn_s);
static unsigned int session_hash(const void* pvdata);
static int session_cmp(const void* pvdata_a, const void* pvdata_b);
static void update_transport_state(enum transport_state trans_state);

/*!
* \brief Allocate memory and initialises the members of the diameter client structure.
*
*/
int diameter_init(struct diameter_cfg* diam_cfg)
{
	if (!diam_cfg)
		return 0;

	if (!(diameter_base = (struct diameter_base*)malloc(sizeof(struct diameter_base))))
    return 0;
	if (!(diameter_base->transport = (struct diameter_peer_transport*)malloc(sizeof(struct diameter_peer_transport)))) {
		ast_log(LOG_ERROR, "Failed to allocate memory for transport\n");
    return 0;
	}

	ast_mutex_init(&diameter_base->baselock);
  diameter_base->config = diam_cfg;
	diameter_base->diameter_thread = AST_PTHREADT_NULL;
	diameter_base->state = CLOSED;
	diameter_base->transactions = NULL;
  diameter_base->transport_timer_id = -1;
	diameter_base->dwr_timerid = -1;
	diameter_base->transport->clientsocket = -1;
	diameter_base->transport->serversocket = -1;
	diameter_base->transport->activesocket = -1;
	
	if (!(diameter_base->scheduler = sched_context_create())) {
		ast_log(LOG_ERROR, "Failed to create scheduler context.\n");
		return 0;
	}

	if (!get_fqdn(&diameter_base->fqdn)) {
		/* The host ip is considered as fqdn if the system api to get hostname fails. */
		if (!(diameter_base->fqdn = (char*)realloc(diameter_base->fqdn, strlen(diameter_base->config->host_ip)+1)))
			return 0;
		strncpy(diameter_base->fqdn, diameter_base->config->host_ip, strlen(diameter_base->config->host_ip)+1);
		diameter_base->fqdn[strlen(diameter_base->config->host_ip)] = '\0';
	}

	/* Hash table for accounting sessions. */
	if (!(diameter_base->acct_sessionlist = ast_hashtab_create(MAX_CALLS, session_cmp, NULL, NULL, session_hash, 1))) {
		ast_log(LOG_ERROR, "Failed to create hash table for accounting sessions.\n");
		return 0;
	}

	if (!diameter_transport_init()) {
		ast_log(LOG_ERROR, "Failed to initialise transport.\n");
		return 0;
	}
	update_transport_state_and_process(TRANSPORT_INIT);

  process_event(START, NULL);
	return 1;
}

/*! \brief Frees the diameter_client structure. */
void diameter_destroy()
{
	if (!diameter_base)
		return;

	if (diameter_base->transport->activesocket > 0) {
		/* Send Disconnect Peer Request to the diameter peer before you stop talking to him. */	
		if (I_OPEN == diameter_base->state || R_OPEN == diameter_base->state) {
			process_event(STOP, NULL);
		}	else {
			take_action(SND_DPR, NULL);
		}
		take_action(CLEANUP, NULL);
	}

	if (diameter_base->acct_sessionlist) {
		ast_hashtab_destroy(diameter_base->acct_sessionlist, cleanup_acct_session);
		diameter_base->acct_sessionlist = NULL;
	}
	if (diameter_base->transactions) {
		transaction_cleanup(&diameter_base->transactions);
		diameter_base->transactions = NULL;
	}

	if (diameter_base->transport)	{
		free(diameter_base->transport);
		diameter_base->transport = NULL;
	}

	ast_mutex_lock(&diameter_base->baselock);

	if (diameter_base->scheduler)	{
		if (diameter_base->transport_timer_id > -1)	
			AST_SCHED_DEL(diameter_base->scheduler, diameter_base->transport_timer_id);

		if (diameter_base->dwr_timerid > -1) 
			AST_SCHED_DEL(diameter_base->scheduler, diameter_base->dwr_timerid);

		sched_context_destroy(diameter_base->scheduler);
		diameter_base->scheduler = NULL;
	}
	ast_mutex_unlock(&diameter_base->baselock);

	if (diameter_base->config) {
		if (diameter_base->config->host_realm) {
			free(diameter_base->config->host_realm);
			diameter_base->config->host_realm = NULL;
		}
		if (diameter_base->config->host_productname) {
			free(diameter_base->config->host_productname);
			diameter_base->config->host_productname = NULL;
		}
		if (diameter_base->config->peer_realm) {
			free(diameter_base->config->peer_realm);
			diameter_base->config->peer_realm = NULL;
		}
		free(diameter_base->config);
		diameter_base->config = NULL;
	}

	if (diameter_base->fqdn) {
		free(diameter_base->fqdn);
    diameter_base->fqdn = NULL;
	}
	
	ast_mutex_destroy(&diameter_base->baselock);

  free(diameter_base);
  diameter_base = NULL;
}

/*!
* \brief Handles CEA message by looking for common applications supported by both the peers and starts a watchdog timer.
*
*/
int handle_CEA(struct message* message_recvd)
{
	struct avp *avp_t = NULL;
	int result = 1;

	if (!(message_recvd && message_recvd->req_avp))	
		return 0;
                
	avp_t = message_recvd->req_avp;

	/* \note Check for other commonly supported applications is not done as Diameter Base accounting is the application currently used.*/
	/* \todo Support for other diameter applications and check for other diameter applications in the CEA received. */

  if (!free_transaction(message_recvd, &diameter_base->transactions))	{
		ast_log(LOG_ERROR, "Failed to free the transaction of CEA message.\n");
		return 0;
	}

	/* Start the watchdog timer to monitor the peer connection */
	ast_mutex_lock(&diameter_base->baselock);
	if (-1 == diameter_base->dwr_timerid)	{
		if (-1 == (diameter_base->dwr_timerid = ast_sched_add(diameter_base->scheduler, 1000*WAIT_TIME, monitor_connection, NULL))) {
			ast_log(LOG_ERROR, "Failed to start the transaction timer.\n");
			result = 0;
		}
	}
	ast_mutex_unlock(&diameter_base->baselock);
	return result;
}

/*!
 * \brief Handles the received CER message.Checks for the commonly supported applications by both the peers.If present a suuccessful
 *  CEA is sent else an unsuccessful CEA(Result Code AVP with value DIAMETER_NO_COMMON_APPLICATION) is sent.
 *
 */
int handle_CER(struct message* message_recvd)
{
	int res = 1;

	if (!message_recvd)
		return 0;

	/* Check if the CER has mandatory Avps. */
	if (1 == (res = check_avp_presence(&message_recvd->req_avp, AVP_HOST_IP_ADDRESS)))
		if (1 == (res = check_avp_presence(&message_recvd->req_avp, AVP_VENDOR_ID)))
	  	res = check_avp_presence(&message_recvd->req_avp, AVP_PRODUCT_NAME);
		
	if (!res)	{
		/* If any of the mandatory Avps is missing send the error result and return. */	
  	return handle_error_message(DIAMETER_MISSING_AVP, message_recvd);
 	}

	/* \note Only base accounting is supported, so not checking for other application .*/
	if (!build_diameter_response(MSG_CER, message_recvd, 0)) {
		ast_log(LOG_ERROR, "Failed to send CEA response.\n");
		return 0;
	}

	/* Register the watchdog timer to monitor the peer connection */
	ast_mutex_lock(&diameter_base->baselock);
	if (-1 == diameter_base->dwr_timerid) {
		if (-1 == (diameter_base->dwr_timerid = ast_sched_add(diameter_base->scheduler, 1000*WAIT_TIME, monitor_connection, NULL))) {
    	ast_log(LOG_ERROR, "Failed to start the transaction timer.\n");
			res = 0;
    }
	}
	ast_mutex_unlock(&diameter_base->baselock);
  return res;
}

/*! \brief Callback function for sending DWR at regular intervals. */ 
int monitor_connection(const void *data)
{
	struct message* msg = NULL;
	int ret_val = 0;

	if (-1 < diameter_base->dwr_timerid) {
		if (1 == (ret_val = build_diameter_request(MSG_DWR, &msg))) 
			ret_val = process_event(SEND_MESSAGE, msg);

		if (!ret_val) {
			ast_log(LOG_WARNING, "Failed to send DWR.\n");
			return -1; 
		}
	}
	return 1;
}	

/*! brief Cleans up the global diameter_client structure for a fresh start.*/
int diameter_cleanup(void) 
{
	if (diameter_base) {
  	diameter_base->transport->disconnected = 0;
    diameter_base->state = CLOSED;

		if (diameter_base->transport->activesocket > 0) {
    	hardclose(diameter_base->transport->activesocket);
      diameter_base->transport->activesocket = -1;
		}
    diameter_base->transport->trans_state = TRANSPORT_CLOSED;

    transaction_cleanup(&diameter_base->transactions);
		diameter_base->transactions = NULL;

		ast_mutex_lock(&diameter_base->baselock);
    if (diameter_base->scheduler) {
			if (diameter_base->transport_timer_id > -1) 
				AST_SCHED_DEL(diameter_base->scheduler, diameter_base->transport_timer_id);

			if (diameter_base->dwr_timerid > -1) 
				AST_SCHED_DEL(diameter_base->scheduler, diameter_base->dwr_timerid);
		}
		ast_mutex_unlock(&diameter_base->baselock);

		if (diameter_base->acct_sessionlist) {
    	ast_hashtab_destroy(diameter_base->acct_sessionlist, cleanup_acct_session);
			diameter_base->acct_sessionlist = NULL;
		}

		if (!(diameter_base->acct_sessionlist = ast_hashtab_create(MAX_CALLS, session_cmp, NULL, NULL, session_hash, 1))) {
			ast_log(LOG_ERROR, "Failed to create hash table for accounting sessions.\n");
      return 0;
		}
		diameter_base->transport->trans_state = TRANSPORT_INIT;
	}	
	return 1;
}

/*! \brief Change the transport state to a new state and take action accordingly. */
int update_transport_state_and_process(enum transport_state trans_state)
{
	enum peer_event_t peer_event;
	int event_exists = 0;

	switch (trans_state) {
	case TRANSPORT_CLOSED:
		diameter_transport_cleanup();
		update_transport_state(trans_state);		
		break;

	case TRANSPORT_INIT:
		update_transport_state(trans_state);		
		break;	

	case TRANSPORT_CONNECTING:
		update_transport_state(trans_state);		
		break;	
			
	case TRANSPORT_RECONNECT:
		if (TRANSPORT_CONNECTING == diameter_base->transport->trans_state) {
			/* still connecting, not yet recieved ack,  the event is I_RCV_CONN_NACK. */
			peer_event = I_RCV_CONN_NACK;
			event_exists = 1;
	    update_transport_state_and_process(TRANSPORT_INIT);
		} else 		 
			update_transport_state(trans_state);		
		break;

	case TRANSPORT_UP:
		if (0 == diameter_base->conn_initiator) { /* When you are responder of the existing connection. */
			update_transport_state(trans_state);		
			break;
		}

		/* When the transport goes up it means connection has been accepted and the event is I_RCV_CONN_ACK. */
		if (TRANSPORT_CONNECTING == diameter_base->transport->trans_state) {
			peer_event = I_RCV_CONN_ACK;
			event_exists = 1;
		}
		update_transport_state(trans_state);		
		break;

	case TRANSPORT_DOWN:
			ast_debug(5, "Transport state machine received transport down while it was in state:%d\n", diameter_base->transport->trans_state);
			update_transport_state(trans_state);		
			if (0 == transport_restart())
				return 0;
			break;	

		default:
			return 0;
	}
	/* If event exists process the existing event. */
	if (event_exists) {
		if (!process_event(peer_event, NULL)) {
			ast_log(LOG_ERROR, "Failed to establish transport connections.\n");
			return 0;
		}
	}
	return 1;	
}

/*! \brief Gets the hostname of the host machine.
 *  \return 1 on success 0 on failure
 */
static int get_fqdn(char** fqdn_s)
{
  char name[128] = "\0";
  char* fqdn = NULL;
  struct hostent *phostent = NULL;
  struct ast_hostent hostent;

  if (-1 == (gethostname(name, sizeof(name))))
    return 0;

  if (!(phostent = ast_gethostbyname(name, &hostent)))
    return 0;

  if (!(fqdn = (char*)malloc(strlen(phostent->h_name)+1)))
    return 0;

  strcpy(fqdn, phostent->h_name);
  *fqdn_s = fqdn;

  return 1;
}

/*!
 * \brief Compare the entries in the hash table
 */
static int session_cmp(const void* pvdata_a, const void* pvdata_b)
{
  if (!(pvdata_a && pvdata_b))
    return -1;
  return(compare_session(pvdata_a, pvdata_b));
}

/*!
 * \brief Hash function for the hash table used for sessions
 */
static unsigned int session_hash(const void* pvdata)
{
  if (!pvdata)
    return -1;
  return (get_session_hash(pvdata));
}

/*! \brief Change the transport state to a new state. */
static void update_transport_state(enum transport_state trans_state)
{
	ast_mutex_lock(&diameter_base->baselock);
	diameter_base->transport->trans_state = trans_state;
	ast_mutex_unlock(&diameter_base->baselock);
}
