/*Copyright and Author
 *
 *This program is free software, distributed under the terms of the GNU General Public License Version 2.
 *
 * \author Leena <leena_m@spanservices.com>, (SPAN Infotech India Pvt. Ltd., Bangalore)
 *
 * \module diameter
 * 
 * \created 04\03\09
 */

/*! \file
  \brief Contains diameter client information.
  
 */

#ifndef _DIAMETER_BASE_H
#define _DIAMETER_BASE_H


#include "diameter_session.h"
#include "diameter_transport.h"

#if defined(__cplusplus) || defined(c_plusplus)
extern "C" {
#endif

/*! \brief Interval used for all the timers used(transport reconnect timer, watchdog timer and transaction timer for requests).*/
#define WAIT_TIME 30

/*! \brief Diameter configuation values */
struct diameter_cfg {
	int 		host_port;           	/*!< TCP port of the diameter client 	*/
	int 		peer_port;           	/*!< port no. of the diameter server 	*/
	int 		buffer_space_avail;  	/*!< buffer space used for storing accounting records (NOT _USED)*/
	char*   host_realm;        		/*!< Realm of the diameter client 		*/
	char*   host_productname; 	 	/*!< Vendor assigned name of the product */
	char* 	peer_realm;        		/*!< realm of the diameter server 		*/
	char 		host_ip[16];         	/*!< ip address of the host machine		*/
	char 		peer_ip[16];         	/*!< ip address of the diameter server */
};


/*! \brief Diameter applications which use this implementation need to register the following callbacks */
/*  \note Is kept for future use and is not used anywhere. */
struct application {
	enum application_name application_identifier; 			/*! Application identifier */
	int (*report_event)(int application_id, int event); /*! call back function to report events */
	struct application* next;
};

/*! \brief Global Information used by the Diameter client */
struct diameter_base {
	int		conn_initiator;															/*!< Indicates whether you are initiator or responder of the connection(1 if initiator and 0 if responder. */
	int		dwr_timerid;																/*!< Timer id of the Watchdog timer */
	int   transport_timer_id;													/*!< Timer id of the transport timer */
	enum peer_state_t       state;                  	/*!< Current state of the diameter client */
 	enum application_name   application;           		/*!< Diameter application supported by both the peers*/
 	pthread_t       diameter_thread;         					/*!< Thread to listen to diameter packets */
	ast_mutex_t 		baselock;
 	char*          	fqdn;                    					/*!< FQDN of the host */	
 	struct diameter_cfg*						config;           /*!< Configuration info */
 	struct sched_context*						scheduler;        /*!< Scheduler for scheduling events at regular intervals of time */
  struct transaction_t*						transactions;     /*!< transactions of CER, DWR, DPR */
	struct ast_hashtab* 						acct_sessionlist; /*!< Hash table for accounting sessions */
	struct diameter_peer_transport* transport;			 	/*!< Connection info */ 
 	/*struct application*	application_registered; */	/*!< List of applications supported */
};

/*! \global variable for diameter client structure.*/
extern struct diameter_base *diameter_base;

/*! 
 * /brief Allocates memory and initialises the members of the global diameter_base structure 
 *
 * /param diam_conf configuration details read from the configuration file 
 *
 * \retval 1 on success and 0 on failure
 */
int diameter_init(struct diameter_cfg *diam_conf);

/*!
 * /brief Frees and removes up the global diameter_base structure
 */
void diameter_destroy(void);

/*!
 * /brief Cleans up the global diameter_base structure for a fresh start( or restart)
 * 
 * \retval 1 on success and 0 on failure
 */
int diameter_cleanup(void);

/*!
 * /brief Processes the received CEA message.
 *
 * /param message_recvd the recived CEA message from the peer
 *
 * \retval 1 on success and 0 on failure
 */
int handle_CEA(struct message* message_recvd);

/*!
 * /brief Processes the received CER message and responds by sending CEA response to the peer.
 *
 * /param message_recvd the recived CER message from the peer
 *
 * \retval 1 on success and 0 on failure
 */
int handle_CER(struct message* message_recvd);

/*!
 * /brief Callback function function for watchdog timer.Sends DWR at regular intervals.
 *
 * /param pvdata is NULL
 */
int monitor_connection(const void *pvdata);
/*!
 * /brief Updated the transport state and takes the corresponding action.
 *
 * /param trans_state state to which the transport is updated to. 
 *
 * \retval 1 on success and 0 on failure
 */
int update_transport_state_and_process(enum transport_state trans_state);

#if defined(__cplusplus) || defined(c_plusplus)
}
#endif

#endif /* _DIAMETER_BASE_H */
