/*Copyright and Author
 *
 *This program is free software, distributed under the terms of the GNU General Public License Version 2.
 *
 * \author Leena <leena_m@spanservices.com>, (SPAN Infotech India Pvt. Ltd., Bangalore)
 *
 * \module diameter
 * 
 * \created 12\03\09
 */

/*! \file
  \brief Diameter message related functions.
  
 */

#include "asterisk.h"	
#include "asterisk/utils.h"
#include "asterisk/sched.h"	
#include "diameter_base.h"
#include "diameter_transaction.h"

#define VER_SIZE                        1       /*!< Version size of a Diameter message */
#define MESSAGE_LENGTH_SIZE             3       /*!< Message length size of a Diameter message */
#define FLAGS_SIZE                      1       /*!< Flags size of a Diameter message */
#define COMMAND_CODE_SIZE               3       /*!< Command code size of a Diameter message */
#define APPLICATION_ID_SIZE             4       /*!< Application Id size of a Diameter message */
#define HOP_BY_HOP_IDENTIFIER_SIZE      4       /*!< Hop-by-Hop Id size of a Diameter message */
#define END_TO_END_IDENTIFIER_SIZE      4       /*!< End-to-End Id size of a Diameter message */
#define AVP_VENDOR_ID_SIZE              4       /*!< Vendor-Id size of an Avp message */

#define AVP_CODE_SIZE      4  	/*!< Code size of the AVP */
#define AVP_FLAGS_SIZE     1  	/*!< Flags size of the AVP */
#define AVP_LENGTH_SIZE    3    /*!< Length size of the AVP */
#define AVP_VENDOR_ID_SIZE 4  	/*!< Vendor Id size of the AVP */

/*! Size of the Diameter Message Header */
#define DIAMETER_MSG_HDR_SIZE  \
        (VER_SIZE + MESSAGE_LENGTH_SIZE + FLAGS_SIZE + COMMAND_CODE_SIZE +\
        APPLICATION_ID_SIZE+HOP_BY_HOP_IDENTIFIER_SIZE+END_TO_END_IDENTIFIER_SIZE)

/*! Size of the AVP Header */
#define AVP_HDR_SIZE(_flags_)  \
        (AVP_CODE_SIZE+AVP_FLAGS_SIZE+AVP_LENGTH_SIZE+\
        AVP_VENDOR_ID_SIZE*(((_flags_)&AVP_FLAG_VENDOR_SPECIFIC)!=0) )

/*! The version field of the Diameter header */
#define MESSAGE_VERSION 1

#define IP_ADDRESS 6	/* no. of bytes for IP address with the family(IPV4) */
#define UNSIGNED32 4	/* no. of bytes for data if Unsigned32 type */
#define UNSIGNED64 8	/* no. of bytes for data if Unsigned64 type */ 
#define to_32x_len( _len_ )  ( (_len_)+(((_len_)&3)?4-((_len_)&3):0) )

/* Private functions */
static int add_req_avps(struct message *diameter_message);
static unsigned int get_three_bytes(unsigned char** stream);
static unsigned int get_four_bytes(unsigned char** stream);
static void set_three_bytes(unsigned char** stream, int int_val);
static void set_four_bytes(unsigned char** stream, int int_val);
static void set_eight_bytes(unsigned char** stream, int int_val);
static int  parse_diameter_message(unsigned char* achbuf, unsigned int buflen, struct message** message_t);


/*! \brief Builds a Diameter request of a specific message code. */
int build_diameter_request(enum msg_code message_code, struct message **msg)
{
	struct message *message_to_build = NULL;

	/*This function builds diameter CER, DPR, DWR ACR requests*/
	if (!(message_to_build = (struct message*)malloc(sizeof(struct message)))) 
		return 0;

	message_to_build->req_avp = NULL;
	message_to_build->command_code = message_code;
	message_to_build->version = MESSAGE_VERSION;

	/* The R bit of the flag is set indicating the message is a Request */
	message_to_build->flags = 0x80;

	/* Set to 0 for CER,DWR and DPR and to 3 if ACR */
	if (MSG_ACR != message_code)
		message_to_build->application_id = DIAMETER_COMMON_MESSAGES;
	else
		message_to_build->application_id = DIAMETER_BASE_ACCOUNTING;

	message_to_build->end_to_end_id = ast_random();
	message_to_build->hop_by_hop_id = ast_random();

	/* Build and add the required avps of the message. */
	if (!add_req_avps(message_to_build)) {
		ast_log(LOG_ERROR, "Failed to add the required AVPs in the request message.\n");
		free_message(message_to_build);
		return 0;
	}
	*msg = message_to_build;
	return 1;
}
	
/*! \brief Builds an ACR request of Record-Type as specified by the record parameter */
int build_diameter_acr_request(void *session_t, enum avp_acct_record record)
{
	char 	avp_val[32] = "\0";
	struct message *acr_message = NULL;
	struct timeval currtime;
	struct avp* avp_t = NULL;
	struct avp* pcur = NULL;
	struct avp* ptemp = NULL;
	long int answertime = 0;
	
	if (!session_t)
		return 0;

	if (!build_diameter_request(MSG_ACR, &acr_message)) {
		ast_log(LOG_ERROR, "Failed to build an ACR message.\n");
		return 0;
	}

	if (!(avp_t = get_session_avp(session_t)))
		ast_log(LOG_ERROR, "avp_t is NULL after get_session_avp.\n");

	/* Adding the mandatory AVPs for accouting */
	if (!build_mandatory_avps(MSG_ACR, &(acr_message->req_avp)))
		goto error;

	/* \todo INTERIM records has to be incremented for subsequent ACR */
	sprintf(avp_val, "%d", get_session_interim_val(session_t));
	if (!build_and_add_avp(AVP_ACCT_RECORD_NUMBER, avp_val, strlen(avp_val), &(acr_message->req_avp), VENDOR_CODE, 0))
		goto error;

	sprintf(avp_val, "%d", record);
	if (!build_and_add_avp(AVP_ACCT_RECORD_TYPE, avp_val, strlen(avp_val), &(acr_message->req_avp), VENDOR_CODE, 0))
		goto error;
	
	/* For Interim records the billable seconds avp is updated */
	/* \note not tested */
	if (AVP_ACCT_INTERIM_RECORD == record) {
		currtime = ast_tvnow();	
		while (NULL != avp_t) {
			if (AVP_AST_BILL_SEC == avp_t->code) {
				answertime = get_session_answertime(session_t);	
				sprintf(avp_val, "%ld", (currtime.tv_sec) - answertime);

				if (!(avp_t->data = (char*)realloc(avp_t->data, strlen(avp_val)+1)))
					goto error;
				strcpy(avp_t->data, avp_val);
				break;
			}
			avp_t = avp_t->next;
		}
	}

	ptemp = acr_message->req_avp;
	if (NULL != (pcur =  get_session_avp_clone(session_t))) {
		while (ptemp->next)
				ptemp = ptemp->next;
			ptemp->next = pcur;
	}

	if (!transaction_send_message(acr_message, transaction_init(acr_message, session_t)))
		goto error;
	return 1;

error:
	free_message(acr_message);
	return 0;
}

/*!
* \brief Builds a diameter response of a specific message code and sends. 
* \note avp_val has value only when an Error Result Code AVP is sent in the response message built else is 0.
*/
int build_diameter_response(enum msg_code message_code, struct message *req_msg, enum diameter_msg_result_code avp_code)
{
	struct message *res_message = NULL;
	enum diameter_msg_result_code res_code = DIAMETER_SUCCESS;
	char avp_val[32] = "\0";
	int res = 1;

	if (!req_msg)
		return 0;

	if (!(res_message = (struct message*)malloc(sizeof(struct message))))
		return 0;

	res_message->req_avp = NULL;
	res_message->application_id = req_msg->application_id;
	res_message->version = MESSAGE_VERSION;

	/* Response message*/
	res_message->flags = 0x00;

	/*Error bit set if error */
	if (0 != avp_code)
		res_message->flags = res_message->flags | 0x20;

	/* Set the end-to-end-id and the hop-by-hop-id same as the response's request. */
	if (req_msg->end_to_end_id)
		res_message->end_to_end_id = req_msg->end_to_end_id;

	if (req_msg->hop_by_hop_id)
		res_message->hop_by_hop_id = req_msg->hop_by_hop_id;

	res_message->command_code = message_code;

	/* Set the Result-Code Avp in the response. For successful response the value is DIAMETER_SUCCESS and error responses the 
		 result code is set accordingly. */
	if (0 != avp_code)
		res_code = avp_code;

	sprintf(avp_val, "%d", res_code); 
	if (!(res = build_and_add_avp(AVP_RESULT_CODE, avp_val, strlen(avp_val), &(res_message->req_avp), VENDOR_CODE, 0))) {
  	ast_log(LOG_ERROR, "Failed to build the Result-Code AVP.\n");		
		goto error_free;
	}

	if (!(res = add_req_avps(res_message))) {
		ast_log(LOG_ERROR, "Failed to add the required AVPs in the response message.\n");
		goto error_free;
	}
	
	if (I_OPEN == diameter_base->state || R_OPEN == diameter_base->state) {
  	if (!(res = process_event(SEND_MESSAGE, res_message)))
			goto error_free;
	} else {
		if(!(res = transaction_send_message(res_message, NULL)))
			goto error_free;
		}

error_free:
	free_message(res_message);
	return res;
}

/*!
* \brief Convert the Diameter message to a buffer to be sent.
*/
int message_to_buffer_nsend(struct message *diameter_message)
{
	unsigned char* buf = NULL; 
	unsigned char send_buf[MAX_BUFFER_SIZE]; 
	unsigned int data_len = 0;
	struct avp* avp = NULL;

	if (!diameter_message || !diameter_message->req_avp)
		return 0;

	memset(send_buf, 0, MAX_BUFFER_SIZE);
	buf = send_buf;

  /* Message Length - For time being set the message length as integer size. The actual length is updated to the header at the end of function*/
	((unsigned int*)buf)[0] = htonl(sizeof(int));

	/* Diameter Version */
	*buf = diameter_message->version;
	buf += VER_SIZE + MESSAGE_LENGTH_SIZE;

	/* Command code */
	((unsigned int*)buf)[0] = htonl(diameter_message->command_code);

	/* Flags */
	*buf = diameter_message->flags;
	buf += FLAGS_SIZE + COMMAND_CODE_SIZE;

	/* Application-ID */
	((unsigned int*)buf)[0] = htonl(diameter_message->application_id);
	buf += APPLICATION_ID_SIZE;

	/* hop by hop id */
	((unsigned int*)buf)[0] = diameter_message->hop_by_hop_id;
	buf += HOP_BY_HOP_IDENTIFIER_SIZE;

	/* end to end id */
	((unsigned int*)buf)[0] = diameter_message->end_to_end_id;
	buf += END_TO_END_IDENTIFIER_SIZE;

	avp = diameter_message->req_avp;
	while (NULL != avp ) {
		set_four_bytes(&buf,avp->code);
		/* flags*/
		(*buf++) = (unsigned char)avp->flag;

		/* avp length */
		switch (avp->type) {
		case AVP_ADDRESS_TYPE:
			data_len = AVP_HDR_SIZE(avp->flag) + IP_ADDRESS;
			break;

		case AVP_UTF8STRING_TYPE:
		case AVP_GROUPED_TYPE:
		case AVP_STRING_TYPE:    
			data_len = AVP_HDR_SIZE(avp->flag) + strlen(avp->data);
			break;

		case AVP_UNSIGNED32_TYPE:
		case AVP_INTEGER32_TYPE: 
			data_len = AVP_HDR_SIZE(avp->flag) + UNSIGNED32;
			break;

		case AVP_UNSIGNED64_TYPE:
			data_len = AVP_HDR_SIZE(avp->flag) + UNSIGNED64;
			break;

		case AVP_ENUMERATED_TYPE:
			data_len = AVP_HDR_SIZE(avp->flag) + 4;
			break;
				
		default:
			data_len = 0;                 
			break;
		}
		if (data_len) {
			set_three_bytes(&buf, data_len);
			data_len = 0;
		}

		/* vendor id */
		if ((avp->flag&0x80)!=0) {
			set_four_bytes(&buf ,avp->vendor_id);
		}

		switch (avp->type) {
			case AVP_ADDRESS_TYPE: 
				((unsigned int*)buf)[0] = htons(1);/* IPV4 Family */
				buf +=2;
				((unsigned int*)buf)[0] = inet_addr(avp->data);
				buf +=6;
				break;

			case AVP_UNSIGNED32_TYPE:
			case AVP_INTEGER32_TYPE: 
				((unsigned int*)buf)[0] = htonl(atoi(avp->data));
				buf += 4;
				break;

			case AVP_UNSIGNED64_TYPE:
				set_eight_bytes(&buf, atoi(avp->data));
				break;

			case AVP_ENUMERATED_TYPE:
				((unsigned int*)buf)[0] = htonl(atoi(avp->data));
				buf += 4;
				break;

			case AVP_UTF8STRING_TYPE:
			case AVP_GROUPED_TYPE:
			case AVP_STRING_TYPE:   
				memcpy(buf, avp->data, strlen(avp->data));
				buf += to_32x_len(strlen(avp->data));
				break;

			default	:             
				break;
		}
		avp = avp->next;
	}
	/* Message Length */
	data_len = buf - send_buf;
	((unsigned int*)send_buf)[0] = htonl(data_len);

	/* Diameter Version */
	*send_buf = diameter_message->version;

	return diameter_transport_send(send_buf, data_len);
}

/*!
* \brief Form a diameter message from the recived buffer and validate the message.
*/
int process_diameter_message(unsigned char* achbuffer, unsigned int buflen)
{
	struct message *message_t = NULL;
	char *avpval = NULL;
	enum peer_event_t peer_event = 0;
	int res = 0;
	int start_dwrtimer = 0;

	if (!(message_t = (struct message*)malloc(sizeof(struct message)))) 
		return 0;
	message_t->req_avp = NULL;

	if (!(res = parse_diameter_message(achbuffer, buflen, &message_t))) {
		goto Cleanup;  
	}

	/* Events are set based on whether i am initiator or responder of the existing connection. */
	if ((message_t->command_code != MSG_CER) && (message_t->command_code != MSG_DWR) && (message_t->command_code != MSG_DPR)) {
		peer_event = (diameter_base->conn_initiator) ? (I_RCV_MESSAGE) : (R_RCV_MESSAGE);

		if (!(res = process_event(peer_event, message_t))) {
			ast_log(LOG_ERROR, "Failed to process RCV_MESSAGE event.\n");
		}
		else
			start_dwrtimer = 1;
		goto Cleanup;
		return 1;
	} else {
		/* Check if it is response message */
		if (NULL != (avpval = get_avpval(&(message_t->req_avp), AVP_RESULT_CODE))) {
			/* Check if the received response is a successful response */
			if ((atoi(avpval) >= 2000) || (atoi(avpval) < 3000)) {
				ast_debug(5, "The message received is a successful response.\n");
				
				/* Process the received response */
				if (find_transaction(message_t, diameter_base->transactions)) {
					switch (message_t->command_code) {
						case MSG_CER:
							peer_event = (diameter_base->conn_initiator) ? (I_RCV_CEA) : (R_RCV_CEA);	
							break;

						case MSG_DWR:
							peer_event = (diameter_base->conn_initiator) ? (I_RCV_DWA) : (R_RCV_DWA);
							break;

						case MSG_DPR:
							peer_event = (diameter_base->conn_initiator) ? (I_RCV_DPA) : (R_RCV_DPA);
							break;
	
						default: 
							res = handle_error_message(DIAMETER_COMMAND_UNSUPPORTED, message_t);
							goto Cleanup;
							
					}/* end of switch */

					res = process_event(peer_event, message_t);
					if(res && (MSG_DWR == message_t->command_code))
						start_dwrtimer = 1;

				} else {
					ast_log(LOG_WARNING, "The recieved message does not belong to any transaction.\n");
				}
			} else {
				/* Check if it is an error response */
				if ((atoi(avpval) >= 3000) && (atoi(avpval) < 6000)) {
					ast_log(LOG_ERROR, "Error message received.\n");
					res = handle_error_message(atoi(avpval), message_t);
				}
			}
			goto Cleanup;
		}

		ast_debug(5, "The message received is a \"request\".\n");

		/* Check if the received request is for local consumption and is to be processed or not */
		/* Check if Destination-Host AVP if present in the request contains the local host's identity */
		if (NULL != (avpval = get_avpval(&(message_t->req_avp), AVP_DESTINATION_HOST))) {
			if (strcmp(avpval, diameter_base->fqdn)) {
				ast_log(LOG_WARNING, "The destination host avp's value does not match the local host's identity and the request is not processed.\n");
        		goto Cleanup;
			}
			/* Check if Destination-Realm AVP if present in the request contains the local host's realm */
		} else if (NULL != (avpval = get_avpval(&(message_t->req_avp), AVP_DESTINATION_REALM))) {
			if (strcmp(avpval, diameter_base->config->host_realm)) {
      	ast_log(LOG_NOTICE, "The destination realm avp's value does not match the local realm and the request is not processed.\n");
        goto Cleanup;
      }
		}

		if (!message_t->req_avp) 
			goto Cleanup;

		/* Check if the received request message has the Orign-host and Origin-realm AVPs in them */
		if (!check_host_nrealm_presence(&message_t->req_avp)) {
			res = handle_error_message(DIAMETER_MISSING_AVP, message_t);
			goto Cleanup;
		}

		switch (message_t->command_code) {
			case MSG_CER:
				if (NULL != (avpval = get_avpval(&message_t->req_avp, AVP_ORIGIN_REALM))) {
        	if (strcmp(avpval, diameter_base->config->peer_realm)) {
						/* if the request's Origin-realm does not match my peer-realm the sender of the request id not 
							 known to me and the connection is closed.*/
          	if (1 == (res = handle_error_message(DIAMETER_UNKNOWN_PEER, message_t))) {
            	if (0 != update_transport_state_and_process(TRANSPORT_DOWN))
								goto Cleanup;
						} 
						else {
							goto Cleanup;
						}
					}
				} else {
					goto Cleanup;
				}
				if(R_OPEN == diameter_base->state || I_OPEN == diameter_base->state)
        	peer_event = R_RCV_CER;
				else
					peer_event = R_CONN_CER;	
        break;

			case MSG_DPR:
				diameter_base->transport->disconnected = 1;
				peer_event = (diameter_base->conn_initiator) ? (I_RCV_DPR) : (R_RCV_DPR);
				break;

			case MSG_DWR:
				peer_event = (diameter_base->conn_initiator) ? (I_RCV_DWR) : (R_RCV_DWR);
				break;

			default:	
				if (!(res = handle_error_message(DIAMETER_COMMAND_UNSUPPORTED, message_t)))
					goto Cleanup;
        }
				/* When a non-CEA message is recieved while waiting for CEA */
				if ((WAIT_I_CEA == diameter_base->state) && (message_t->command_code != MSG_CER)) {
					if (MSG_DPR == message_t->command_code)
					/* When a DPR message is recieved while waiting for CEA */
          	peer_event = I_PEER_DISC;
					else
						peer_event = I_RCV_NON_CEA;
				}
				if (!(res = process_event(peer_event, message_t))) {
      		ast_log(LOG_ERROR, "Failed to process event.\n");
					goto Cleanup;
       	}

				if(MSG_DWR == message_t->command_code) {
  		   	ast_mutex_lock(&diameter_base->baselock);
	    	  if(-1 == (diameter_base->dwr_timerid = ast_sched_replace(diameter_base->dwr_timerid, diameter_base->scheduler, 1000*WAIT_TIME, monitor_connection, NULL))){
     				ast_log(LOG_ERROR, "Failed to reschedule watchdog timer.\n");
       			res = 0;
   				}
    			ast_mutex_unlock(&diameter_base->baselock);    	
					goto Cleanup;
				}				
			}
		res = 1;

Cleanup:
	if(start_dwrtimer) {
		if(diameter_base->dwr_timerid < 0) {
     	ast_mutex_lock(&diameter_base->baselock);
      if(-1 == (diameter_base->dwr_timerid = ast_sched_add(diameter_base->scheduler, 1000*WAIT_TIME, monitor_connection, NULL))){
     		ast_log(LOG_ERROR, "Failed to start the watchdog timer.\n");
       	res = 0;
   		}
    	ast_mutex_unlock(&diameter_base->baselock);    	
  	}
	}
	free_message(message_t);
	return res;
}

/*! \brief Frees a diameter message */
void free_message(struct message *msg)
{
	if (!msg)
		return;

	if (msg->req_avp) {
		avplist_cleanup((msg->req_avp));
		msg->req_avp = NULL;
	}
	free(msg);
	msg = NULL;
}				

/*! \brief To parse the message into 3 bytes */
static unsigned int get_three_bytes(unsigned char** str_data)
{
  unsigned int temp = 0 ;
  long byte_in_stream = 0;
  unsigned char* stream = *str_data;

  byte_in_stream = *stream++;   /* get first byte */
  temp = (unsigned int)byte_in_stream<<16;

  byte_in_stream = *stream++;  /* get second byte */
  temp |= ((unsigned int)byte_in_stream)<<8;

  byte_in_stream = *stream++;    /* get third byte */
  temp |= ((unsigned int)byte_in_stream);

  *str_data = stream;
  return(temp);
}

/*! \brief To parse the message into 4 bytes */
static unsigned int get_four_bytes(unsigned char** str_data)
{
  unsigned int temp = 0 ;
  long byte_in_stream = 0;
  unsigned char* stream = *str_data;

  byte_in_stream = *stream++;   /* get first byte */
  temp = (unsigned int)byte_in_stream << 24;

  byte_in_stream = *stream++;  /* get second byte */
  temp |= ((unsigned int)byte_in_stream)<<16;

  byte_in_stream = *stream++;    /* get third byte */
  temp |= ((unsigned int)byte_in_stream)<<8;

  byte_in_stream = *stream++;    /* get fourth byte */
  temp |= ((unsigned int)byte_in_stream);

  *str_data = stream;
  return(temp);
}

/*! \brief To convert the data into 3 bytes for transmission */
static void set_three_bytes(unsigned char** str_data, int int_val)
{
  unsigned int temp = 0 ;
  unsigned char* stream = *str_data;

  temp = (int_val & 0x00ff0000) >> 16;
  *stream++ = (unsigned char)temp;
  
  temp = (int_val & 0x0000ff00) >> 8;
  *stream++ = (unsigned char)temp;

  temp = (int_val & 0x000000ff);
  *stream++ = (unsigned char)temp;

  *str_data = stream;
}


/*! \brief To convert the data into 4 bytes for transmission */
static void set_four_bytes(unsigned char** str_data, int int_val)
{
  unsigned int temp = 0 ;
  unsigned char* stream = *str_data;
	
  temp = (int_val & 0xff000000)>> 24;
  *stream++ = (unsigned char)temp;

  temp = (int_val & 0x00ff0000) >> 16;
  *stream++ = (unsigned char)temp;
  
  temp = (int_val & 0x0000ff00) >> 8;
  *stream++ = (unsigned char)temp;

  temp = (int_val & 0x000000ff);
  *stream++ = (unsigned char)temp;

  *str_data = stream;
}


/*! \brief To convert the AVP data into 8 bytes for transmission */
static void set_eight_bytes(unsigned char** str_data, int int_val)
{
  unsigned int temp = 0 ;
  unsigned char* stream = *str_data;

  temp = (int_val & 0xff00000000000000ULL)>> 56;
  *stream++ = (unsigned char)temp;

  temp = (int_val & 0x00ff000000000000ULL) >> 48;
  *stream++ = (unsigned char)temp;
  
  temp = (int_val & 0x0000ff0000000000ULL) >> 40;
  *stream++ = (unsigned char)temp;

  temp = (int_val & 0x000000ff00000000ULL) >> 32;
  *stream++ = (unsigned char)temp;

  temp = (int_val & 0x00000000ff000000ULL)>> 24;
  *stream++ = (unsigned char)temp;

  temp = (int_val & 0x0000000000ff0000ULL) >> 16;
  *stream++ = (unsigned char)temp;
  
  temp = (int_val & 0x000000000000ff00ULL) >> 8;
  *stream++ = (unsigned char)temp;

  temp = (int_val & 0x00000000000000ffULL);
  *stream++ = (unsigned char)temp;

  *str_data = stream;
}


/*!
* \brief Add the mandatory AVPs of a Diameter message.
*
* \param diameter_message The message to which the AVPS are to be added
* \retval 1 on success and 0 on failure
*/
static int add_req_avps(struct message *diameter_message)
{
  char avp_val[32] = "\0";
  int result = 1;

  if (!diameter_message)
    return 0;

  if (!build_and_add_avp(AVP_ORIGIN_HOST, diameter_base->fqdn, strlen(diameter_base->fqdn), &(diameter_message->req_avp), VENDOR_CODE, 0)) {
    ast_log(LOG_ERROR, "Failed to build and add Origin-Host Avp\n");
    return 0;
  }

  if (!build_and_add_avp(AVP_ORIGIN_REALM, diameter_base->config->host_realm, strlen(diameter_base->config->host_realm), &(diameter_message->req_avp), VENDOR_CODE, 0)) {
    ast_log(LOG_ERROR, "Failed to build and add Origin-Realm Avp\n");
    return 0;
  }

  switch (diameter_message->command_code) {
    case MSG_CER:
      result = build_mandatory_avps(MSG_CER, &(diameter_message->req_avp));
      break;

    case MSG_DPR:
      if (diameter_message->flags & 0x80) {
        sprintf(avp_val, "%d", DO_NOT_WANT_TO_TALK_TO_YOU);
        result = build_and_add_avp(AVP_DISCONNECT_CAUSE, avp_val, strlen(avp_val), &(diameter_message->req_avp), VENDOR_CODE, 0);
      }
      break;

    default:
      break;

  }/* end of switch */
  return result;
}

/*!
* \brief Convert the received buffer to Diameter message structure.
*
* \param achbuf the received buffer
*
* \return message on success else NULL
*/
static int parse_diameter_message(unsigned char* achbuf, unsigned int buflen, struct message** message)
{
  struct message* message_t = NULL;
  struct avp*     avp_t = NULL;
  unsigned char*  msg_ptr;
  unsigned char   avp_flags;
  unsigned int    message_len = 0;
  unsigned int    avp_len = 0;
  unsigned int    avpdata_len = 0;
  unsigned int    avp_code = 0;
  unsigned int    avp_vendorid = 0;
  int temp = 0;
  char buf[32] = "\0";
  int error = 1;

  if (!achbuf || !buflen || !message || !(*message))
    return 0;

  message_t = *message;
  msg_ptr = achbuf;

  /* Getting the Version */
  message_t->version = (unsigned char)*msg_ptr;
  msg_ptr += VER_SIZE;

  /* Getting Message length */
  if (buflen < (message_len = get_three_bytes(&msg_ptr))) {
    ast_log(LOG_ERROR, "Diameter message len %u bigger then buffer len %u\n",message_len, buflen);
    return 0;
  }

  /* Getting Command flags */
  message_t->flags = *msg_ptr;
  msg_ptr += FLAGS_SIZE;

  /* Getting Command code */
  message_t->command_code = get_three_bytes(&msg_ptr);

  /* Getting Application-Id */
  message_t->application_id = get_four_bytes(&msg_ptr);

  /* Getting Hop-by-Hop-Id */
  message_t->hop_by_hop_id = *((unsigned int*)msg_ptr);
  msg_ptr += HOP_BY_HOP_IDENTIFIER_SIZE;

  /* Getting End-to-End-Id */
  message_t->end_to_end_id = *((unsigned int*)msg_ptr);
  msg_ptr += END_TO_END_IDENTIFIER_SIZE;

	/* diameter version supported is "1" */
  if ((0x80 & message_t->flags) && (message_t->version != 1)) {
    if (!handle_error_message(DIAMETER_UNSUPPORTED_VERSION, message_t)) {
      ast_log(LOG_ERROR, "The message is of version %d which is not valid ", message_t->version);
    }
    return 0;
  }

  /* If the message header size of the received diameter request is not of standard size then throw the error response to the peer */
  if (((msg_ptr - achbuf != DIAMETER_MSG_HDR_SIZE)) && (0x80 & message_t->flags)) {
    ast_debug(1, "Message received with invalid header combination.\n");
    if (!handle_error_message(DIAMETER_INVALID_HDR_BITS, message_t)) {
      ast_log(LOG_ERROR, "Failed to send error response message.\n");
    }
    return 0;
  }

  /* Get the AVPs from the message */
  while (msg_ptr < achbuf + message_len) {
    /*Avp code */
    avp_code = get_four_bytes(&msg_ptr);

    /* Avp flags*/
    avp_flags = *msg_ptr;
    msg_ptr += AVP_FLAGS_SIZE;

    if (0x80 & message_t->flags) {
      switch (avp_flags) {
        case AVP_FLAG_MANDATORY:
        case AVP_FLAG_RESERVED:
        case AVP_FLAG_VENDOR_SPECIFIC:
        case AVP_FLAG_END_TO_END_ENCRYPT:
          break;

        default:
          /* unrecognised avp flags; sending error*/
          if (!handle_error_message(DIAMETER_INVALID_AVP_BITS, message_t)) {
            ast_log(LOG_ERROR, "Failed to send error response message.\n");
          }
          return 0;
      }
    }
    /*Avp length*/
    if ((avp_len = get_three_bytes(&msg_ptr))  < 1) {
      ast_debug(1, "Invalid AVP length \n");
      return 0;
    }

    /* Avp vendor-ID */
    if (avp_flags & 0x80) {
      avp_vendorid = get_four_bytes(&msg_ptr);
    }

    /* Data length*/
    avpdata_len = avp_len - AVP_HDR_SIZE(avp_flags);

    if (achbuf + message_len < msg_ptr + avpdata_len) {
      ast_log(LOG_ERROR,"source buffer too short to read a whole data for AVP!\n");
      return 0;
    }

    /*Create the AVP */
    if (avp_code == AVP_ACCT_RECORD_TYPE || avp_code == AVP_RESULT_CODE || avp_code == AVP_ACCT_RECORD_NUMBER) {
      temp = get_four_bytes(&msg_ptr);
      msg_ptr -= 4;
      sprintf(buf,"%d", temp);
      if (!build_and_add_avp(avp_code, buf, strlen(buf), &(message_t->req_avp), avp_vendorid, avp_flags)) {
        ast_log(LOG_ERROR, "Failed to create and the AVP into the message's list.\n");
        return 0;
      }
    } else {
      if (!build_and_add_avp(avp_code, (char*)msg_ptr, avpdata_len, &(message_t->req_avp), avp_vendorid, avp_flags)) {
        ast_log(LOG_ERROR, "Failed to create and the AVP into the message's list.\n");
        return 0;
      }
    }
    msg_ptr += to_32x_len(avpdata_len);

  }/*end of while*/
  avp_t = message_t->req_avp;
  if (message_t->flags & 0x80) {
    while (NULL != avp_t) {
      if (AVP_FLAG_UNDEFINED == avp_t->flag) {
        error = handle_error_message(DIAMETER_AVP_UNSUPPORTED, message_t);
        break;
      }
      avp_t = avp_t->next;
    }
  }
  return error;
}


