/*Copyright and Author
 *
 *This program is free software, distributed under the terms of the GNU General Public License Version 2.
 *
 * \author Leena <leena_m@spanservices.com>, (SPAN Infotech India Pvt. Ltd., Bangalore)
 *
 * \module diameter
 * 
 * \created 12\03\09
 */

/*! \file
  \brief Diameter message info.
  
 */

#ifndef _DIAMETER_MESSAGE_H
#define _DIAMETER_MESSAGE_H

#include "diameter_avp.h"

#if defined(__cplusplus) || defined(c_plusplus)
extern "C" {
#endif

/*! Messages supported by the Diameter Client */
enum msg_code {
	MSG_CER = 257,
	MSG_ACR = 271,
	MSG_DWR = 280,
	MSG_DPR = 282
};

/*! Applications supported by the Diameter client */
enum application_name {
	DIAMETER_COMMON_MESSAGES	= 0,
	DIAMETER_NASREQ 					= 1,
	DIAMETER_MOBILE_IP 				= 2,
	DIAMETER_BASE_ACCOUNTING 	= 3
};

/*! brief Structure to hold the Diameter message */
struct message {
	int 					version;				/*!< Version field of the Diameter message */
	int 					length;					/*!< Message Length field of the Diameter message */
	unsigned int  command_code;		/*!< Command-Code field of the Diameter message */
 	unsigned int  application_id; /*!< Application-ID field of the Diameter message */
	unsigned int  end_to_end_id;  /*!< End-to-End Identifier field of the Diameter message */
 	unsigned int  hop_by_hop_id;  /*!< Hop-by-Hop Identifier field of the Diameter message */
	struct avp*		req_avp;				/*!< The mandatory AVPs of the Diameter message */
	unsigned char flags;					/*!< Command Flags field of the Diameter message */
};

/*! 
 * \brief Builds a diameter request message.
 *
 * \param message_code the message code of the diameter message to be built 
 * \param msg the built request is assigned to this parameter.
 *
 * \retval 1 on success and 0 on failure
 */
int build_diameter_request(enum msg_code message_code, struct message **msg);

/*! 
 * \brief The received Diameter message is parsed,validated and processed.
 *
 * \param buffer the buffer which contains the received Diameter message
 * \param buflen the length of the param buffer
 *
 * \retval 1 on success and 0 on failure
 */
int process_diameter_message(unsigned char* achbuffer, unsigned int buflen);

/*! 
 * \brief Builds an accounting diameter request message.
 *
 * \param session the accounting session to which the built request belongs
 * \param record the accounting record of the accounting request to be built
 *
 * \retval 1 on success and 0 on failure
 */
int build_diameter_acr_request(void *session_t, enum avp_acct_record record);

/*! 
 * \brief Builds a diameter response message
 *
 * \param message_code message code of the response to be built
 * \param req_msg the request message for which the response is built
 * \param avp_val value of the Result-Code AVP to be set in the response message
 *
 * \retval 1 on success and 0 on failure
 */
int build_diameter_response(enum msg_code message_code, struct message *req_msg, enum diameter_msg_result_code avp_val);

/*!
 * \brief Converts the diameter message to buffer.
 * \param diameter_message the diameter message which is converted to buffer   
 *
 * \retval 1 on success and 0 on failure
 */
int message_to_buffer_nsend(struct message *diameter_message);

/*!
* \brief Frees the message.
*
* \param msg the message to be freed
*/
void free_message(struct message *msg);

#if defined(__cplusplus) || defined(c_plusplus)
}
#endif

#endif /*_DIAMETER_MESSAGE_H */
