/*Copyright and Author
 *
 *This program is free software, distributed under the terms of the GNU General Public License Version 2.
 *
 * \author Leena <leena_m@spanservices.com>, (SPAN Infotech India Pvt. Ltd., Bangalore)
 *
 * \module diameter
 * 
 * \created 01\04\09
 */

/*! \file
  \brief Contains accounting session related routines and peer state handling routines.
  
 */

#include "asterisk.h"
#include "asterisk/utils.h"
#include "asterisk/hashtab.h"
#include "asterisk/sched.h"
#include "diameter_base.h"
#include "diameter_transaction.h"

/* function declaration */
static int election(struct message *msg);
static int handle_message(struct message* msg);
static int failure_tosend_check(struct message* msg);
static int transmit_acr_interimrequest(const void* pvdata);
static enum acct_event_t event_if_realtime_avp(struct message *msg);
static int handle_ACA(struct message *aca_message, struct acct_session *session);
static int find_session_ntrans(struct message *msg, struct acct_session **session);
static int acct_sessionstate_process(struct acct_session* acct_session_t, enum acct_event_t acct_event);
static enum acct_event_t validate_aca_avps(struct message* msg, int realtime_flag, enum avp_acct_record rec_t);

/*!
 * \brief Send the ACR of a user of the specified record
 *
 * \param avplist list of asterisk specific(which contain cdr info)AVPs to be sent in the ACR to be sent
 * \param rec_type Record-Type of the Accounting request
 *
 * \retval 1 on success and 0 n failure
 */
int send_accounting_request(struct avp* avplist, enum avp_acct_record rec_type)
{
	struct acct_session *session_t = NULL;
	struct acct_session tmp;
	enum acct_event_t acct_event = ACCT_REQUEST_ACCESS;
	char* avp_answer_time = NULL;
	char* avp_session_id = NULL;
	
	if (!avplist)
		return 0;

	if ((AVP_ACCT_START_RECORD == rec_type) || (AVP_ACCT_EVENT_RECORD == rec_type)) {
		if(!(avp_answer_time = get_avpval(&(avplist), AVP_AST_ANSWER_TIME)))
			return 0;

		if (!(avp_session_id = get_avpval(&(avplist), AVP_SESSION_ID)))
			return 0;

		if (!(session_t = (struct acct_session*)malloc(sizeof(struct acct_session))))
			return 0;

		session_t->answer_time = atoi(avp_answer_time);
		session_t->session_id = atoi(avp_session_id);
		session_t->state = IDLE;
		session_t->interim_val = 0;
		session_t->interim_time_val = 0;
		session_t->avp_toadd = avplist;
		ast_mutex_init(&session_t->mutex);
		session_t->acct_timer_id = -1;
		session_t->trans = NULL;

		/* add into the hash table */
		ast_mutex_lock(&diameter_base->baselock);
		if (!ast_hashtab_insert_safe(diameter_base->acct_sessionlist, session_t)) {
			ast_debug(1, "The entry %d is already in table!\n",session_t->session_id);
			cleanup_acct_session(session_t);
			ast_mutex_unlock(&diameter_base->baselock);
			return 0;
		}
		ast_mutex_unlock(&diameter_base->baselock);

		if (AVP_ACCT_EVENT_RECORD == rec_type)
			acct_event = ACCT_ONE_TIME_SERVICE;
	} else if (AVP_ACCT_STOP_RECORD == rec_type) {/* look for the user's session */
		if (!(avp_session_id = get_avpval(&(avplist), AVP_SESSION_ID)))
		return 0;

		tmp.session_id = atoi(avp_session_id);

		if (!(session_t = ast_hashtab_lookup(diameter_base->acct_sessionlist, &tmp)))
			return 0;

		acct_event = ACCT_USR_SERVICE_TERMINATED;
	}

	if (!acct_sessionstate_process(session_t, acct_event)) {
		ast_mutex_lock(&diameter_base->baselock);
		ast_hashtab_remove_this_object(diameter_base->acct_sessionlist, session_t);
		ast_mutex_unlock(&diameter_base->baselock);
		cleanup_acct_session(session_t);
		return 0;
	}
	return 1;
}

/*!
 * \brief Frees the accounting session.
 *
 * \param pvdata session to be freed
 *
 * \retval 1 on success and 0 on failure
 */
void cleanup_acct_session(void* pvdata)
{
	struct acct_session *session_t = (struct acct_session*) pvdata;

	if (!session_t)
		return;

	transaction_cleanup(&session_t->trans);
	session_t->trans = NULL;

	avplist_cleanup((session_t->avp_toadd));

	ast_mutex_destroy(&(session_t->mutex));
	free(session_t);
	session_t = NULL;
}


/*! \brief Compare the entries in the hash table. */
int compare_session(const struct acct_session* session_a, const struct acct_session* session_b)
{
	return session_a->session_id != session_b->session_id; /* return 0 on a match */
}

/*! \brief Hash function for the hash table used for sessions. */
unsigned int get_session_hash(const struct acct_session* session)
{
	return session->session_id;
}

/*!
 * \brief Based on the action the corresponding functions are called
 *
 * \param action The action to be taken
 * \param msg message needed to perform the action if present else NULL
 *
 * \note msg will be present only if the message has caused the action else will be NULL
 * \retval 1 on success and 0 on failure 
 */
int take_action(enum action_t action, struct message* msg)
{
	enum peer_event_t peer_event = I_RCV_CONN_ACK;
	struct message *req_msg = NULL;
	enum transport_state trans_state = TRANSPORT_CLOSED;
	int result = 1;

	switch (action) {
		case SND_CONN_REQ:
			trans_state = diameter_transport_connection();
			if (TRANSPORT_CONNECTING == trans_state) {
      	ast_debug(1, "diameter_transport_connection() is trying to connect\n");
				break;
			} else if(TRANSPORT_CLOSED == trans_state) {
				ast_debug(1, "diameter_transport_connection() failed.\n");
				peer_event = I_RCV_CONN_NACK;
			}
			if(!(result = process_event(peer_event, NULL)))
				ast_log(LOG_ERROR, "Failed to establish transport connections.\n");
			break;

		case ACCEPT:
			/* The connection is already accepted */
			break;

		case SND_CER:
			if (!(result = build_diameter_request(MSG_CER, &req_msg))) {
				ast_log(LOG_ERROR, "Failed to build CER.\n");
				break;
			}

			if (!(result = transaction_send_message(req_msg, transaction_init(req_msg, NULL)))) {
				free_message(req_msg);				
			}
			break;

		case PROCESS_CER:
			if (!(result = (msg && handle_CER(msg)))) {
				ast_log(LOG_ERROR, "Failed to handle CER.\n");
			}
			break;

		case PROCESS_CEA:
			if (!(result = (msg && handle_CEA(msg)))) {
				ast_log(LOG_ERROR, "Failed to handle CEA.\n");
			}
			break;

		case PROCESS_DWR:
			if (!(result = (msg && build_diameter_response(MSG_DWR, msg, 0)))) {
				ast_log(LOG_ERROR, "Failed to build message DWA\n");
			}
			break;

		case PROCESS_DWA:
			if (!(result = (msg && free_transaction(msg, &diameter_base->transactions)))) {
				ast_log(LOG_ERROR, "Failed to free the transaction of DWA message.\n");
			}
			break;

		case SND_DPR:
			if (!(result = build_diameter_request(MSG_DPR, &req_msg))) {
      	ast_log(LOG_ERROR, "Failed to build DPR.\n");
				break;
			}

			if (!(result = transaction_send_message(req_msg, transaction_init(req_msg, NULL)))) {
				free_message(req_msg);
			}
			break;

		case SND_MESSAGE:
			if (msg && (msg->flags & 0x80)) {
				result = transaction_send_message(msg, transaction_init(msg, NULL)); 
			} else {
				result = transaction_send_message(msg, NULL);
			}
			break;

		case SND_DPA:
			if (!(result = (msg && build_diameter_response(MSG_DPR, msg, 0)))) {
				ast_log(LOG_ERROR, "Failed to build message DPA\n");
     	}
   		break;

		/* Not used */
		case REJECT:
			/* Do Nothing */
			break;

		case ERROR:
		case I_DISC:	
		case R_DISC:
			/* The transport is restarted */
			if (!(result = update_transport_state_and_process(TRANSPORT_DOWN))) {
      	ast_log(LOG_ERROR, "Failed to restart the transport\n");
			}
			break;

		case CLEANUP:
			if (!(result = update_transport_state_and_process(TRANSPORT_CLOSED))) {
      	ast_log(LOG_ERROR, "Failed to restart the transport\n");
			}
			break;
			
		case PROCESS:
			if (!(result = (msg && handle_message(msg)))) {
   			ast_log(LOG_ERROR, "Failed to handle message\n");
   		}
			break;

		default:     
			break;
	}
	return result;
}

/*! \brief Update the diameter client state */
void update_diameterbase_state(enum peer_state_t state)
{
	ast_mutex_lock(&diameter_base->baselock);
	diameter_base->state = state;
	ast_mutex_unlock(&diameter_base->baselock);
}
	
/*!
 * \brief Based on the event that takes place the corresponding actions are called
 *
 * \param peer_event Event that has taken place
 * \param msg message that caused the event
 *
 * \retval 1 on success and 0 on failure 
 */
int process_event(enum peer_event_t peer_event, struct message* msg)
{
	int result = 1;

	switch (diameter_base->state) {
		case CLOSED:
			switch (peer_event) {
				case START:
					update_diameterbase_state(WAIT_CONN_ACK);
					if (!(result = take_action(SND_CONN_REQ, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"SND_CONN_REQ\"\n");
					}
					break;

				case R_CONN_CER:
					if (1 == (result = take_action(ACCEPT, msg)))
						if (1 == (result = take_action(PROCESS_CER, msg))) {
							update_diameterbase_state(R_OPEN);
						}
						break;

				default:
					break;
			}
			break;

		case WAIT_CONN_ACK:
			switch (peer_event) {
				case I_RCV_CONN_ACK:
					if (!(result = take_action(SND_CER, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"SND_CER\"\n");
						break;
					}
					update_diameterbase_state(WAIT_I_CEA);
					break;

				case I_RCV_CONN_NACK:
					if (-1 == diameter_base->transport_timer_id) {
						if (!(result = register_transporttimer())) {	
							ast_log(LOG_ERROR, "Failed to register transport timer.\n");
							break;
						}
					}				
					update_diameterbase_state(CLOSED);
					break;
				
				/* not reachable */
				case R_CONN_CER:
          if (!(result = take_action(ACCEPT, NULL))) {
          	ast_log(LOG_ERROR, "Failed to take action \"ACCEPT\"\n");
						break;
					}
          if (take_action(PROCESS_CER, msg))
						update_diameterbase_state(WAIT_CONN_ACK_ELECT);
					else
						update_diameterbase_state(WAIT_CONN_ACK);
					break;

				case TIMEOUT:
					update_diameterbase_state(CLOSED);
					if (!(result = take_action(ERROR, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"ERROR\"\n");
					}
					break;

				default:
					ast_debug(5, "Event not handled in state WAIT_CONN_ACK.\n");
					break;
			}
			break;

		case WAIT_I_CEA:
			switch (peer_event) {
				case I_RCV_CEA:
					if (!(result = take_action(PROCESS_CEA, msg))) {
						if (!(result = take_action(R_DISC, NULL))) {
							ast_log(LOG_ERROR, "Failed to take action \"DISC\"\n");
						}
						break;
					}
					update_diameterbase_state(I_OPEN);
					break;

				case R_CONN_CER:
					if (!(result = take_action(ACCEPT, NULL))) {
          	ast_log(LOG_ERROR, "Failed to take action \"ACCEPT\"\n");
						break;
   				}

  				if (!(result = take_action(PROCESS_CER, msg))) {
						ast_log(LOG_ERROR, "Failed to take action \"PROCESS_CER\"\n");
						break;
   				}
					update_diameterbase_state(WAIT_RETURNS);
					if (election(msg))
						process_event(WIN_ELECTION, msg);
					break;

				case I_PEER_DISC:
					if (!(result = take_action(I_DISC, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"I_DISC\"\n");
						break;
					}
					update_diameterbase_state(CLOSED);
					break;

				case I_RCV_NON_CEA:
				case TIMEOUT:
					update_diameterbase_state(CLOSED);
					if (!(result = take_action(ERROR, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"ERROR\"\n");
					}
					break;

				default:
					ast_debug(5, "Event not handled in WAIT_I_CEA.\n");
					break;
			}
			break;

		/* not reachable */
		case WAIT_CONN_ACK_ELECT:
			switch (peer_event) {
				default:
					break;
			}
			break;

		case WAIT_RETURNS:
			switch (peer_event) {
				case WIN_ELECTION:
				case I_PEER_DISC:
					/* The condition will not be met as only one connection is supported */
					if (!(result = take_action(I_DISC, NULL))) {
          	ast_log(LOG_ERROR, "Failed to take action \"I_DISC\"\n");
						break;
					}
					if (take_action(PROCESS_CER, msg))
						update_diameterbase_state(R_OPEN);
					else {
						if (!(result = take_action(R_DISC, NULL))) {
							ast_log(LOG_ERROR, "Failed to take action \"R_DISC\"\n");
							break;
						}
					}
					update_diameterbase_state(CLOSED);
					break;

				case I_RCV_CEA:
					if (!(result = take_action(R_DISC, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"R_DISC\"\n");
						break;
					}
					if (!(result = take_action(PROCESS_CEA, msg))) {
						ast_log(LOG_ERROR, "Failed to PROCESS_CEA.\n");
						break;
					}
					update_diameterbase_state(I_OPEN);
					break;

				case R_PEER_DISC:
					if (!(result = take_action(R_DISC, NULL))) {
          	ast_log(LOG_ERROR, "Failed to take action \"R_DISC\"\n");
						break;
					}
					update_diameterbase_state(WAIT_I_CEA);
					break;

				case R_CONN_CER:
					if (!(result = take_action(REJECT, NULL))) {
          	ast_log(LOG_ERROR, "Failed to take action \"REJECT\"\n");
						break;
					}
					update_diameterbase_state(WAIT_RETURNS);
					break;

				case TIMEOUT:
					if (!(result = take_action(ERROR, NULL))) {
          	ast_log(LOG_ERROR, "Failed to take action \"ERROR\"\n");
						break;						
					}
					update_diameterbase_state(CLOSED);
					break;	

				default:
					ast_debug(5, "Event not handled in state WAIT_RETURNS.\n");
					break;
			}
			break;

		case R_OPEN:
			switch (peer_event) {
				case SEND_MESSAGE:
					if (!(result = take_action(SND_MESSAGE, msg))) {
						ast_log(LOG_ERROR, "Failed to take action \"SND_MESSAGE\"\n");
					}
					break;

				case R_RCV_MESSAGE:
					if (!(result = take_action(PROCESS, msg))) {
						ast_log(LOG_ERROR, "Failed to take action \"PROCESS\"\n");
					}
					break;

				case R_RCV_DWR:
					if (!(result = take_action(PROCESS_DWR, msg))) {
						ast_log(LOG_ERROR, "Failed to take action \"PROCESS_DWR\"\n");
					}
					break;

				case R_RCV_DWA: 
					if (!(result = take_action(PROCESS_DWA, msg))) {
						ast_log(LOG_ERROR, "Failed to take action \"PROCESS_DWA\"\n");
					}
					break;

				/* Unreachable code as only one active connection is supported */
				case R_CONN_CER:
					if (!(result = take_action(REJECT, NULL))) {
          	ast_log(LOG_ERROR, "Failed to take action \"REJECT\"\n");
					}
					break;

				case STOP:
					if (!(result = take_action(SND_DPR, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"SND_DPR\"\n");
						break;
					}
					update_diameterbase_state(CLOSING);
					break;

				case R_RCV_DPR:
					if (!(result = (msg && msg->req_avp)))
  						break;

					if (!(result = check_avp_presence(&msg->req_avp, AVP_DISCONNECT_CAUSE)))
						break;

					if (!(result = take_action(SND_DPA, msg))) {
						ast_log(LOG_ERROR, "Failed to take action \"SND_DPA\"\n");
						break;
					}
				  take_action(R_DISC, NULL);
					update_diameterbase_state(CLOSED);
					break;

				/* Not Used */	
				case R_PEER_DISC:	
					if (!(result = take_action(R_DISC, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"R_PEER_DISC\"\n");
						break;
					}
					update_diameterbase_state(CLOSED);
					break;

				case R_RCV_CER:
          if (take_action(PROCESS_CER, msg))
						update_diameterbase_state(R_OPEN);
					else {
						if (!(result = take_action(R_DISC, NULL))) {
							ast_log(LOG_ERROR, "Failed to take action \"R_DISC\"\n");
							break;
						}
						update_diameterbase_state(CLOSED);
					}
					break;

				case R_RCV_CEA:
          if (!(result = take_action(PROCESS_CEA, msg))) {
						if (!(result = take_action(R_DISC, NULL))) {
							ast_log(LOG_ERROR, "Failed to take action \"R_DISC\"\n");
							break;
						}
						update_diameterbase_state(CLOSED);
						break;
					}
					update_diameterbase_state(R_OPEN);
					break;

				default:
					ast_debug(5, "Event not handled R_OPEN.\n");
					break;
			}
			break;

		case I_OPEN:
			switch (peer_event) {
				case SEND_MESSAGE:
					if (!(result = take_action(SND_MESSAGE, msg))) {
						ast_log(LOG_ERROR, "Failed to take action \"SND_MESSAGE\"\n");
					}
					break;

				case I_RCV_MESSAGE:
					if (!(result = take_action(PROCESS, msg))) {
						ast_log(LOG_ERROR, "Failed to take action \"PROCESS\"\n");
					}
					break;

				case I_RCV_DWR:
					if (!(result = take_action(PROCESS_DWR, msg))) {
						ast_log(LOG_ERROR, "Failed to take action \"PROCESS_DWR\"\n");
					}
					break;

				case I_RCV_DWA:
					if (!(result = take_action(PROCESS_DWA, msg))) {
						ast_log(LOG_ERROR, "Failed to take action \"PROCESS_DWA\"\n");
					}
					break;

				/* not reachable */
				case R_CONN_CER:
					if (!(result = take_action(REJECT, NULL))) {
        		ast_log(LOG_ERROR, "Failed to take action \"R_CONN_CER\"\n");
					}
					break;
					
				case STOP:
					if (!(result = take_action(SND_DPR, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"STOP\"\n");
						break;
					}
					update_diameterbase_state(CLOSING);
					break;

				case I_RCV_DPR:
					if (!(result = (msg && msg->req_avp)))
							break;

					if (!(result = check_avp_presence(&msg->req_avp, AVP_DISCONNECT_CAUSE)))
							break;

					if (!(result = take_action(SND_DPA, msg))) {
						ast_log(LOG_ERROR, "Failed to take action \"SND_DPA\"\n");
						break;
					}
					take_action(I_DISC, NULL);
					update_diameterbase_state(CLOSED);
					break;

				/* not used */
				case I_PEER_DISC:
					if (!(result = take_action(SND_DPA, msg))) {
						ast_log(LOG_ERROR, "Failed to take action \"I_PEER_DISC\"\n");
						break;
					}
					if (!(result = take_action(I_DISC, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"I_DISC\"\n");
						break;
					}
					update_diameterbase_state(CLOSED);
					break;

				/* not reachable */		
				case I_RCV_CER:
					if (take_action(PROCESS_CER, msg))
						update_diameterbase_state(R_OPEN);
					else {
						if (!(result = take_action(R_DISC, NULL))) {
							ast_log(LOG_ERROR, "Failed to take action \"R_DISC\"\n");
							break;
						}
						update_diameterbase_state(CLOSED);
					}
        	break;

				case I_RCV_CEA:
					if (!(result = take_action(PROCESS_CEA, msg))) {
						if (!(result = take_action(I_DISC, NULL))) {
							ast_log(LOG_ERROR, "Failed to take action \"I_DISC\"\n");
							break;
						}
						update_diameterbase_state(CLOSED);
						break;
					}
					update_diameterbase_state(I_OPEN);
					break;

				default:
					ast_debug(5, "Event not handled in I_OPEN.\n");
					break;
			}
			break;

		case CLOSING:
			switch (peer_event) {
				case I_RCV_DPA:
					/* not used as disconnection is taken care */
					/*if(!(result = take_action(I_DISC, NULL))){
							ast_log(LOG_ERROR, "Failed to take action \"I_DISC\"\n");
							break;
					}*/
					update_diameterbase_state(CLOSED);
					break;

				case R_RCV_DPA:
					/* not used as disconnection is taken care */
					/*if(!(result = take_action(R_DISC, NULL))){
						ast_log(LOG_ERROR, "Failed to take action \"R_DISC\"\n");
						break;
					}*/
					update_diameterbase_state(CLOSED);
					break;

				case TIMEOUT:
					update_diameterbase_state(CLOSED);
					if (!(result = take_action(ERROR, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"ERROR\"\n");
					}
					break;

				case I_PEER_DISC:
					if (!(result = take_action(I_DISC, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"I_DISC\"\n");
						break;
					}
					update_diameterbase_state(CLOSED);
					break;

				case R_PEER_DISC:
					if (!(result = take_action(R_DISC, NULL))) {
						ast_log(LOG_ERROR, "Failed to take action \"R_DISC\"\n");
						break;
					}
					update_diameterbase_state(CLOSED);
          break;

				default:
					ast_debug(5, "Event not handled in CLOSING.\n");
					break;
			}
			break;

		default:
			ast_log(LOG_NOTICE, "Event not handled for current state.\n");
			break;

	} /* end of switch */
	return result;
}

/*! \brief Gets the cdrinfo avplist of the session */
struct avp* get_session_avp(struct acct_session* session)
{
	if (session)
	 	return session->avp_toadd;

	return NULL;
}

/*! \brief Gets the value of the number of interim record of the session */
    /*note not used */
int get_session_interim_val(struct acct_session* session)
{
	if (session)
		return session->interim_val;

	return 0;
}

/*! \brief Gets the answertime of the call of the session */
long int get_session_answertime(struct acct_session* session)
{
	if (session)
		return session->answer_time;

	return 0;
}

struct avp* get_session_avp_clone(struct acct_session *session)
{
	struct avp* avplist_clone = NULL;
	struct avp* pcur = NULL;
	
	if (!(session && session->avp_toadd))
		return NULL;

	pcur = session->avp_toadd;

	while (NULL != pcur) {
		if (!build_and_add_avp(pcur->code, pcur->data, strlen(pcur->data), &avplist_clone, pcur->vendor_id, pcur->flag)) {
			avplist_cleanup(avplist_clone);			
			return NULL;
		}
		pcur = pcur->next;
	}
	return avplist_clone;
}


static int election(struct message *msg)
{
  char* avpval = NULL;
  char* str_c = diameter_base->fqdn;
  char* str_s = NULL;
  int len = 0;
  int i = 0;
  int byte = 0;

  if (!(msg && msg->req_avp))
        return 0;

  if (!(avpval = get_avpval(&(msg->req_avp), AVP_ORIGIN_HOST))) {
    ast_log(LOG_NOTICE, "Failed to get the value of \"AVP_ORIGIN_HOST\" avp.\n");
    return 0;
  }

  str_s = avpval;
  len = ((strlen(str_c) > strlen(str_s))? (strlen(str_c)) : (strlen(str_s)));
  for (i = 0; i < len; len++) {
    byte = ((unsigned char) str_c[i])-((unsigned char) str_s[i]);
    if (byte > 0)
      return 1;
    else if(byte < 0)
      return 0;
  }
  if (strlen(str_c) > strlen(str_s))
    return 1;
  return 0;
}

/*!
 * \brief Search for the session and transaction of an accounting message.
 *
 * \param msg message whose session is to be searched
 *
 * return session param is assigned value(message's session) on success.
 */
static int find_session_ntrans(struct message *msg, struct acct_session **session)
{
  struct acct_session  tmp_sess;
  struct acct_session* res = NULL;
  char* avpval = NULL;

  if (!(msg && msg->req_avp))
    return 0;

  if(!(avpval = get_avpval(&(msg->req_avp), AVP_SESSION_ID)))
    return 0;

  tmp_sess.session_id = atoi(avpval);

  if (NULL != (res = ast_hashtab_lookup(diameter_base->acct_sessionlist, &tmp_sess))) {
    ast_debug(1, "Session found for the message.\n");
    if (find_transaction(msg, res->trans)) {
      *session = res;
      ast_debug(1, "Transaction found the message.\n");
      return 1;
    }
    ast_debug(1, "Failed to find transaction of the accounting message.\n");
  }
  ast_debug(1, "Failed to find session of the accounting message.\n");
  return 0;
}

/*!
 * \brief Handles messages other than CER, CEA, DPR, DPA, DWR or DWA.
 *
 * \param msg message which is to processed.
 *
 * \retval 1 on success and 0 on failure
 */
static int handle_message(struct message *msg)
{
  struct acct_session *session_t = NULL;
  char* avpval = NULL;

  /*\note As of now only Accounting answer message is handled */
  if (!msg || !msg->req_avp || (MSG_ACR != msg->command_code))
    return 0;

  /* check if the message is a response message */
  if (!(avpval = get_avpval(&msg->req_avp, AVP_RESULT_CODE)))
    return 0;

  if (find_session_ntrans(msg, &session_t)) {
    if (NULL != session_t) {
      return handle_ACA(msg, session_t);
 		}
	}
  return 0;
}

/*! \brief Sends error messages, also handles error messages.*/
int handle_error_message(int error_code, struct message *err_msg)
{
  enum action_t action;

  if (!err_msg)
    return 0;


  /* \note As of now only one incoming error message is recognised. */
  if (DIAMETER_NO_COMMON_APPLICATION == error_code) {
    /* The connection is closed when the two Diameter peers do not have a common application */
    ast_log(LOG_NOTICE, "No common application exists between the two peers.\n");
    action = (diameter_base->conn_initiator) ? (I_DISC) : (R_DISC);
    return take_action(action, NULL);
  }
  else {
    return build_diameter_response(err_msg->command_code, err_msg, error_code);
  }

  return 1;
}

/*!
* \brief Callback for interim timer.Sends the interim accounting request when interim interval elapses.
* \note Not used now.
*/
static int transmit_acr_interimrequest(const void* pvdata)
{
  struct acct_session *session = NULL;

  if (!(session = (struct acct_session*)pvdata))
    return 0;

  return acct_sessionstate_process(session, ACCT_INTERIM_INTERVAL_ELAPSES);
}

/*!
* \brief Checks if "Failure to send"(as defined in RFC 3588) is met or not.
*
* \param msg the message received from the peer
*
* \retval 1 if the condition is met else 0
*/
static int failure_tosend_check(struct message* msg)
{
  char* avpval = NULL;

  /* Get the Result-Code Avp value and check for the error codes. */
  if(!(avpval = get_avpval(&(msg->req_avp), AVP_RESULT_CODE)))
    return 0;

  switch (atoi(avpval)) {
    case DIAMETER_OUT_OF_SPACE:
    case DIAMETER_TOO_BUSY:
    case DIAMETER_LOOP_DETECTED:
      return 1;
    default:
      break;
  }
  return 0;
}

/*!
 * \brief Checks if the values of the Acct_realtime_required avp is any of three specific values.
    Also checks for buffer space availability and takes the corresponding action.
 *
 * \param msg the accounting answer received from the peer
 *
 * \retval if any of the conditions is met returns the event else returns 1(means is a suucessful accounting answer)
 */
static enum acct_event_t event_if_realtime_avp(struct message *msg)
{
  enum avp_acct_realtime realtime_avpval = DELIVER_AND_GRANT;
  int avail = diameter_base->config->buffer_space_avail;
  char* avpval = NULL;

  /* \todo Storing of accounting records in buffer.*/
  /* \note As of now the "buffering" variable from configuration file is read "buffer_space_avail" is set high which is used here.*/

  if(!(msg && msg->req_avp))
    return 0;

  if (!(avpval = get_avpval(&msg->req_avp, AVP_ACCT_REALTIME_REQUIRED)))
    return 0;

  realtime_avpval = atoi(avpval);

  if (avail) {
    if (DELIVER_AND_GRANT != realtime_avpval)
      return ACCT_FAILURE_TOSEND_BA_NOT_DELIVER_AND_GRANT;
  } else {
    if (GRANT_AND_LOSE == realtime_avpval)
      return ACCT_FAILURE_TOSEND_NBA_GRANT_NLOSE;
    else
      return ACCT_FAILURE_TOSEND_NBA_NOT_GRANT_NLOSE;
  }
  return 1;
}

/*!
 * \brief Validates the received Accounting answer for various conditions(like realtime, buffer availability) according to their
          record-types and takes the corresponding action.
 *
 * \param msg the accounting message to be validated
 * \param realtime_flag is set high if realtime check has to happen else 0
 * \param rec_t Record-Type of the accounting answer
 *
 * \retval 1 if the accounting answer is successful or returns the event if any of the conditions are met else returns 0
 */
static enum acct_event_t validate_aca_avps(struct message* msg, int realtime_flag, enum avp_acct_record rec_t)
{
  enum acct_event_t acct_event = ACCT_FAILED_START_ANS_RECVD;
  enum avp_acct_record acct_rec_event =AVP_ACCT_START_RECORD;
  char* avpval = NULL;
  int res = 0;

  if(!(msg && msg->req_avp))
    return 0;

  if (!failure_tosend_check(msg)) {
    if (NULL != (avpval = get_avpval(&(msg->req_avp), AVP_RESULT_CODE))) {
       if (!(res = ((atoi(avpval) >= DIAMETER_AVP_UNSUPPORTED) && (atoi(avpval) <= DIAMETER_INVALID_AVP_BIT_COMBO))))
        return 1;
    } else
      return 0;
  }

  switch (rec_t) {
    case AVP_ACCT_START_RECORD:
    case AVP_ACCT_INTERIM_RECORD:
      if (!res) {
        acct_event = event_if_realtime_avp(msg);
      } else {
        if (realtime_flag) {
          if (realtime_avpcheck(msg->req_avp, GRANT_AND_LOSE)) {
            if (AVP_ACCT_START_RECORD == acct_rec_event)
              acct_event = ACCT_FAILED_START_ANS_RECVD;
                  else if (AVP_ACCT_INTERIM_RECORD == acct_rec_event)
              acct_event = ACCT_FAILED_INTERIM_ANS_RCVD_GRANT_NLOSE;
          }
          else {
            if (AVP_ACCT_START_RECORD == acct_rec_event)
              acct_event = ACCT_FAILED_ANSWER_NOT_GRANT_NLOSE;
            else if (AVP_ACCT_INTERIM_RECORD == acct_rec_event)
              acct_event = ACCT_FAILED_INTERIM_ANS_RCVD_NOT_GRANT_NLOSE;
          }
        }
      }
      break;

    case AVP_ACCT_STOP_RECORD:
      ((res)?(acct_event = ACCT_FAILED_STOP_ANS_RECVD):(acct_event = ACCT_FAILURE_TOSEND_BA));
      break;

    case AVP_ACCT_EVENT_RECORD:
      ((res)?(acct_event = ACCT_FAILED_EVENT_ANS_RECVD):(acct_event = ACCT_FAILURE_TOSEND_BA));
      break;
  }
  return acct_event;
}

/*!
 * \brief Handles an Accounting Answer message in accordance with their Record-Type
 *
 * \param aca_message Accounting message received
 * \param session_t Accounting session to which the accounting message(aca_message param) belongs
 *
 * \retval 1 on success and 0 on failure
 */
static int handle_ACA(struct message *aca_message, struct acct_session *session_t)
{
  char* avpval = NULL;
  enum acct_event_t acct_event = 0;

  if (!(aca_message && session_t))
    return 0;

  if (!(avpval = get_avpval(&(aca_message->req_avp), AVP_ACCT_RECORD_TYPE)))
    return 0;

  if(1 == (acct_event = validate_aca_avps(aca_message, 1, atoi(avpval))))
  {
    switch (atoi(avpval)) {
      case AVP_ACCT_START_RECORD:
      /* When accounting interval is sent by the diameter server in the ACA it responds with */
      if (NULL != (avpval = get_avpval(&(aca_message->req_avp), AVP_ACCT_INTERIM_INTERVAL))) {
        ast_mutex_lock(&session_t->mutex);
        session_t->interim_time_val = atoi(avpval);
        ast_mutex_unlock(&session_t->mutex);
      }

      if (0 != session_t->interim_time_val) {
        /* Schedule the interim timer */
        ast_mutex_lock(&diameter_base->baselock);
        if (-1 == (session_t->acct_timer_id = ast_sched_add(diameter_base->scheduler, session_t->interim_time_val, transmit_acr_interimrequest, session_t))) {
          ast_log(LOG_ERROR, "Failed to schedule the interim timer.\n");
          ast_mutex_unlock(&diameter_base->baselock);
          return 0;
        }
        ast_mutex_unlock(&diameter_base->baselock);
      }
      acct_event = ACCT_SUCCESSFUL_START_ANS_RECVD;
    break;

    case AVP_ACCT_INTERIM_RECORD:
        ast_mutex_lock(&session_t->mutex);
        session_t->interim_val++;
        ast_mutex_unlock(&session_t->mutex);
        acct_event = ACCT_SUCCESSFUL_INTERIM_ANS_RECVD;
      break;

    case AVP_ACCT_STOP_RECORD:
      acct_event = ACCT_SUCCESSFUL_STOP_ANS_RECVD;
      break;

    case AVP_ACCT_EVENT_RECORD:
      acct_event = ACCT_SUCCESSFUL_EVENT_ANS_RECVD;
      break;

		default:
        return 0;
    }
  }

  if (!acct_sessionstate_process(session_t, acct_event)) {
    ast_log(LOG_ERROR, "Failed to update the accounting session.\n");
    return 0;
  }

  if (session_t && session_t->trans) {
    if (!free_transaction(aca_message, &session_t->trans)) {
      ast_log(LOG_ERROR, "Failed to free the transaction in session %d\n",session_t->session_id);
      return 0;
    }
  }
  return 1;
}

/*!
 * \brief Updates the state of an Accounting session
 *
 * \param acct_session_t session whose state is to be updated
 * \event acct_event event that took place and triggered the call
 *
 * \retval 1 on success and 0 on failure
 */
static int acct_sessionstate_process(struct acct_session *acct_session_t, enum acct_event_t acct_event)
{
  enum acct_event_t event_s = 0;

  if (!acct_session_t)
    return 0;

  ast_mutex_lock(&acct_session_t->mutex);

  switch (acct_session_t->state) {
  case IDLE:
    switch (acct_event) {
    case ACCT_REQUEST_ACCESS:
      event_s = AVP_ACCT_START_RECORD;
      break;

    case ACCT_ONE_TIME_SERVICE:
      event_s = AVP_ACCT_EVENT_RECORD;
      break;

    case ACCT_RECORDS_IN_STORAGE:
      /* \todo Store records in buffer */
      break;

    default:
      break;
    }

    if (AVP_ACCT_START_RECORD == event_s || AVP_ACCT_EVENT_RECORD == event_s) {
      if (!build_diameter_acr_request(acct_session_t, event_s)) {
        ast_log(LOG_ERROR, "Failed to build accounting request of in session %d\n", acct_session_t->session_id);
        ast_mutex_unlock(&acct_session_t->mutex);
        return 0;
      }

      if (AVP_ACCT_START_RECORD != event_s)
        acct_session_t->state = PENDING_E;
      else
        acct_session_t->state = PENDING_S;
    }
    break;

  case PENDING_S:
    switch (acct_event) {
    case ACCT_SUCCESSFUL_START_ANS_RECVD:
    case ACCT_FAILURE_TOSEND_NBA_GRANT_NLOSE:
    case ACCT_FAILURE_TOSEND_NBA_NOT_GRANT_NLOSE:
    case ACCT_FAILED_START_ANS_RECVD:
    case ACCT_FAILURE_TOSEND_BA_NOT_DELIVER_AND_GRANT:
      acct_session_t->state = OPEN;
      break;

    case ACCT_FAILED_ANSWER_NOT_GRANT_NLOSE:
      /* disconnecting the user */
      acct_session_t->state = IDLE;
      goto remove_session;

    case ACCT_USR_SERVICE_TERMINATED:
      /* \todo Store Stop Record */
      break;
    default:
      break;
    }
    break;

    case OPEN:
      switch (acct_event) {
        case ACCT_INTERIM_INTERVAL_ELAPSES:
          event_s = AVP_ACCT_INTERIM_RECORD;
          break;

        case ACCT_USR_SERVICE_TERMINATED:
          event_s = AVP_ACCT_STOP_RECORD;
          break;

        default:
          break;
      }
      if (AVP_ACCT_INTERIM_RECORD == event_s || AVP_ACCT_STOP_RECORD == event_s) {
        if (!build_diameter_acr_request(acct_session_t, event_s)) {
          ast_log(LOG_ERROR, "Failed to build accounting request of START_RECORD in session %d\n", acct_session_t->session_id);
          ast_mutex_unlock(&acct_session_t->mutex);
          return 0;
        }
        if (AVP_ACCT_INTERIM_RECORD != event_s)
          acct_session_t->state = PENDING_L; /*This is stop record event*/
        else
          acct_session_t->state = PENDING_I;
      }
      break;

    case PENDING_I:
      switch (acct_event) {
        case ACCT_SUCCESSFUL_INTERIM_ANS_RECVD:
        case ACCT_FAILURE_TOSEND_NBA_GRANT_NLOSE:
        case ACCT_FAILED_INTERIM_ANS_RCVD_GRANT_NLOSE:
        case ACCT_FAILURE_TOSEND_BA_NOT_DELIVER_AND_GRANT:
          acct_session_t->state = OPEN;
          break;

        case ACCT_FAILURE_TOSEND_NBA_NOT_GRANT_NLOSE:
          /* \todo Disconnect the user */
        case ACCT_FAILED_INTERIM_ANS_RCVD_NOT_GRANT_NLOSE:
          /* \todo Disconnect the user */
        case ACCT_USR_SERVICE_TERMINATED:
          /* \todo Store the stop record */

          /* state is retained in PENDING_I */
          break;

        default:
          break;
      }
      break;

    case PENDING_E:
      switch (acct_event) {
        case ACCT_SUCCESSFUL_EVENT_ANS_RECVD:
          acct_session_t->state = IDLE;
          goto remove_session;

        case ACCT_FAILURE_TOSEND_BA:
          /* todo Store event record */

        case ACCT_FAILURE_TOSEND_NBA:
        case ACCT_FAILED_EVENT_ANS_RECVD:
          acct_session_t->state = IDLE;
          break;
        default:
          break;
      }
      break;

    /* this state is never reached */
    case PENDING_B:
      switch (acct_event) {
        case ACCT_SUCCESSFUL_ACCT_ANS_RCVD:
        case ACCT_FAILED_ACCT_ANS_RCVD:
          acct_session_t->state = IDLE;
          goto remove_session;

        case ACCT_FAILURE_TOSEND:
          acct_session_t->state = IDLE;
          break;
        default:
          break;
      }
      break;

    case PENDING_L:
      switch (acct_event) {
        case ACCT_SUCCESSFUL_STOP_ANS_RECVD:
          acct_session_t->state = IDLE;
          goto remove_session;

        case ACCT_FAILURE_TOSEND_BA:
          /* \todo Store stop record */
        case ACCT_FAILURE_TOSEND_NBA:
        case ACCT_FAILED_STOP_ANS_RECVD:
          acct_session_t->state = IDLE;
          break;
        default:
          break;
      }
      default:
        break;
  }/* end of switch */

  ast_mutex_unlock(&acct_session_t->mutex);
  return 1;

remove_session:
  ast_hashtab_remove_this_object(diameter_base->acct_sessionlist, acct_session_t);
  ast_mutex_unlock(&acct_session_t->mutex);
  cleanup_acct_session(acct_session_t);
  ast_mutex_unlock(&acct_session_t->mutex);
  return 1;
}

