/*Copyright and Author
 *
 *This program is free software, distributed under the terms of the GNU General Public License Version 2.
 *
 * \author Leena <leena_m@spanservices.com>, (SPAN Infotech India Pvt. Ltd., Bangalore)
 *
 * \module diameter
 * 
 * \created 10\04\09
 */

/*! \file
  \brief Transaction related routines.
  
 */

#include "asterisk.h"
#include "asterisk/utils.h"
#include "asterisk/hashtab.h"
#include "asterisk/sched.h"
#include "diameter_base.h"
#include "diameter_transaction.h"

static int retransmit_request(const void* data);

/*! \brief Create a transaction structure for a Diameter message. */
struct transaction_t* transaction_init(struct message* diam_message, void* pvdata)
{
	struct transaction_t* trans = NULL;
	struct acct_session* session = (struct acct_session*)pvdata;

	if (!diam_message) 
		return NULL;

	if (!(trans = (struct transaction_t*)malloc(sizeof(struct transaction_t)))) 
		return NULL;

	if (diam_message->end_to_end_id)
		trans->endtoend_identifier = diam_message->end_to_end_id;

	if (diam_message->hop_by_hop_id)
		trans->hopbyhop_identifier = diam_message->hop_by_hop_id;

	trans->request_message = diam_message;
	trans->retransmissions = 0;
	trans->timer_id = -1;
	trans->next = NULL;

	/*add the newly created transaction to the list */
	if (!session) {
 		ast_mutex_lock(&diameter_base->baselock);
  	if (!diameter_base->transactions) {
    	trans->next = diameter_base->transactions;
    }
   	diameter_base->transactions = trans;
    ast_mutex_unlock(&diameter_base->baselock);
	} else {
  	ast_mutex_lock(&session->mutex);
   	if (!session->trans) {
    	trans->next = session->trans;
  	}
  	session->trans= trans;
    ast_mutex_unlock(&session->mutex);
 	}
	return trans;
}

int find_transaction(struct message *msg, struct transaction_t* translist)
{
	struct transaction_t* transaction_list = NULL;

	if (!( msg && translist))
		return 0;

	transaction_list = translist;

	while (NULL != transaction_list) {
		if (msg->hop_by_hop_id == transaction_list->request_message->hop_by_hop_id) {
			return 1; /* found transaction */
		}
		transaction_list = transaction_list->next;
	}
	ast_debug(5, "Transaction not found for the message.\n");
	return 0;
}


/*! \brief Free the transaction. */
int free_transaction(struct message *message, struct transaction_t** trans)
{
	struct transaction_t *trans_prev = NULL;
	struct transaction_t *trans_curr = NULL;

	if (!(message && trans))
		return 0;

	trans_curr = trans_prev = *trans;

	while (NULL != trans_curr) {
		if (message->hop_by_hop_id == trans_curr->request_message->hop_by_hop_id) {
			trans_prev->next = trans_curr->next;
			if (trans_prev == trans_curr)
      	*trans = NULL;
			trans_curr->next = NULL;
   		transaction_cleanup(&trans_curr);
			return 1;
 		}
		trans_prev = trans_curr;
		trans_curr = trans_curr->next;
	}
	return 0;
}

int transaction_send_message(struct message *message_to_send, struct transaction_t* transaction)
{
	if (!message_to_send)
		return 0;

	if ((0x80 & message_to_send->flags) && !transaction)
		return 0;

	if (!message_to_buffer_nsend(message_to_send)) 
 		return 0;

	/* Start the transaction timer */
	if (transaction) {
		ast_mutex_lock(&diameter_base->baselock);
    if (-1 == (transaction->timer_id = ast_sched_add(diameter_base->scheduler, 1000*WAIT_TIME, retransmit_request, transaction))){
			ast_log(LOG_ERROR, "Failed to start the transaction timer.\n");
  		ast_mutex_unlock(&diameter_base->baselock);
   		return 0;
  	}

		if (diameter_base->dwr_timerid > 0 && message_to_send->command_code != MSG_DWR)
			AST_SCHED_DEL(diameter_base->scheduler, diameter_base->dwr_timerid);

		ast_mutex_unlock(&diameter_base->baselock);
	}
	return 1;
}

/*! \brief Cleans up a transaction list */
void transaction_cleanup(struct transaction_t** trans_t)
{
	struct transaction_t* cur_trans = NULL;
	struct transaction_t* trans = NULL;

	if (!trans_t)
		return;

	cur_trans = *trans_t;

	while (NULL != cur_trans) {
		trans = cur_trans;
		cur_trans = cur_trans->next;

		if (trans->timer_id > -1) {
			ast_mutex_lock(&diameter_base->baselock);
			AST_SCHED_DEL(diameter_base->scheduler, trans->timer_id);
  		ast_mutex_unlock(&diameter_base->baselock);
 		}
 
		if (trans->request_message) 
			free_message(trans->request_message);

   	free(trans);
  	trans = NULL;
	}
}

/*! \brief Retransmit the request when no response is recieved after 30 seconds and if request DWR the function restarts the transport. */
static int retransmit_request(const void* pvdata)
{
	struct transaction_t* transaction = NULL;
	struct message *msg = NULL;
	struct acct_session tmp;
	struct acct_session* session_t = NULL;
	char* sessionid = NULL;
	enum msg_code messagecode;

	if (!(transaction = (struct transaction_t*)pvdata))    
		return -1;

	if(!transaction->request_message)
		return -1;

	messagecode = transaction->request_message->command_code;

	if (MSG_ACR == messagecode) {
		if (NULL != (sessionid = get_avpval(&transaction->request_message->req_avp, AVP_SESSION_ID))) {
   		tmp.session_id = atoi(sessionid);
 			if (!(session_t = ast_hashtab_lookup(diameter_base->acct_sessionlist, &tmp))) {
 				return -1;
			}
  	}
	}

	if (messagecode != MSG_DWR) {
		if (transaction->retransmissions < MAX_RETRANSMISSIONS) {
			if (MSG_ACR == messagecode)
				ast_mutex_lock(&session_t->mutex);

			/* Setting the T flag high for retransmission */
			transaction->request_message->flags = (transaction->request_message->flags | 0x10);
  		if (!message_to_buffer_nsend(transaction->request_message)) {
    		if (MSG_ACR == messagecode)
      		ast_mutex_unlock(&session_t->mutex);
        return -1;
      }
  		transaction->retransmissions++;
     	if (MSG_ACR == messagecode)
      	ast_mutex_unlock(&session_t->mutex);
      return 1;
 		} else {
 			/* Send DWR request after MAX_RETRANSMISSIONS */
 			if (!build_diameter_request(MSG_DWR, &msg)) {
     		ast_log(LOG_ERROR, "Failed to build DWR.\n");
   			return -1;
   		}

  		if (!process_event(SEND_MESSAGE, msg)) {
   			ast_log(LOG_ERROR, "Failed to process the event \"SND_MESSAGE\"\n");
      	return -1;
   		}

    	if (MSG_ACR == messagecode) {
     		ast_hashtab_remove_this_object(diameter_base->acct_sessionlist, session_t);
     		cleanup_acct_session(session_t);
    	}
     	return 0;
		}
	}

	if ((diameter_base->state == I_OPEN) || (diameter_base->state == R_OPEN)) 
		update_diameterbase_state(CLOSING);

	/*scheduler timer id for this transaction is setting to -1. This is because we cannot delete this scheduled job from scheduler in transaction cleanup while restarting the transport. The return 0 of this function automatically delete the scheduled job from scheduler*/
	transaction->timer_id = -1;
	if (!process_event(TIMEOUT, NULL)) {
		ast_log(LOG_ERROR, "Failed to process event TIMEOUT.\n");
 	}
	return 0;
}

