/*Copyright and Author
 *
 *This program is free software, distributed under the terms of the GNU General Public License Version 2.
 *
 * \author Leena <leena_m@spanservices.com>, (SPAN Infotech India Pvt. Ltd., Bangalore)
 *
 * \module diameter
 * 
 * \created 24\03\09
 */

/*! \file
  \brief Connection related information.
  
 */

#include "asterisk.h"
#include "asterisk/sched.h"
#include "asterisk/utils.h"
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include "diameter_base.h"

#define DIAMETER_PORT 3868
#define INVALID_SOCKET -1


AST_MUTEX_DEFINE_STATIC(schedlock);	/*!< Protecting the schedule_event thread */

/* Private functions */
static int create_endpoint(void);
static int create_new_socket(void);
static int restart_connecting_socket(void);
static int re_connect(const void *data);
static void* transport_tcp_listener(void* pvdata);
static const char* transport_error_string(int err_no);


/*!
* \brief Creates sockets for to establish connection with the peer.
*
* \retval 1 on success and 0 on failure 
*/
int diameter_transport_init()
{
	diameter_base->transport->disconnected = 0;
  diameter_base->transport->trans_state = TRANSPORT_CLOSED;

	/* create socket for making a connection with the server */
	if (!create_endpoint()) {
		ast_log(LOG_ERROR, "Failed to create sockets for connection.\n");
		return 0;
	}	

	/* Be ready to accept incoming connections */
	/* make the server listening socket as non-blocking */
	if (-1 == listen(diameter_base->transport->serversocket, 5)) {
		ast_log(LOG_ERROR, "listen error is [%s]\n", transport_error_string(errno));	
		return 0;
	}

	/* Thread to keep listening for incoming Diameter packetes */
	if (ast_pthread_create(&diameter_base->diameter_thread, NULL, transport_tcp_listener, NULL)) {
		ast_log(LOG_ERROR, "Failed to create diameter transport TCPlistener thread\n");
		return 0;
	}
	return 1;
}


/*!
* \brief Sends message to the diameter peer(server).
*
* \param data data to be sent
* \param data_length length of the data to sent
*
* \retval 1 on success and 0 on failure 
*/
int diameter_transport_send(const unsigned char *data, int data_length)
{
	if (!data)
		return 0;

	if (TRANSPORT_UP == diameter_base->transport->trans_state) {
		if (-1 == send(diameter_base->transport->activesocket, data, data_length, 0)) {
			ast_log(LOG_ERROR, "send error is [%s]\n", transport_error_string(errno));	
			return 0;
		}
	}
	return 1;
}

/*! brief Makes connection attempts with the peer */
enum transport_state diameter_transport_connection()
{
	struct sockaddr_in server_addr;
	enum transport_state trans_state = TRANSPORT_CLOSED;
	
  server_addr.sin_family = AF_INET;
  server_addr.sin_addr.s_addr = inet_addr(diameter_base->config->peer_ip);
  server_addr.sin_port = htons(diameter_base->config->peer_port);
	
	/* Receive oobdata inline */
	setoobinline(diameter_base->transport->clientsocket);
	
	if (TRANSPORT_INIT == diameter_base->transport->trans_state) {
		if (-1 == connect(diameter_base->transport->clientsocket, (struct sockaddr *)&server_addr, sizeof(struct sockaddr))) {
			if (EINPROGRESS != errno) {			
				/* Connect failed */
				ast_log(LOG_ERROR, "Connect error is [%s]\n", transport_error_string(errno));	
				return TRANSPORT_CLOSED;
			} else { 	
				/* Unblocked connection */
				trans_state = TRANSPORT_CONNECTING;
				
			}
		} else {
			/* Connect successful */
			diameter_base->transport->activesocket = diameter_base->transport->clientsocket;
			diameter_base->conn_initiator = 1;	
			trans_state = TRANSPORT_UP;
		}
		update_transport_state_and_process(trans_state);
	}
	return diameter_base->transport->trans_state;
}

/*!
* \brief Kills the listener thread and cleans up the socket
*
* \retval 1 on success and 0 on failure
*/
void diameter_transport_cleanup()
{
  /* Only if there is any connection alive cleanup */
  if (diameter_base->transport->trans_state >= TRANSPORT_INIT) {
    if (diameter_base->transport->clientsocket > 0) {
      hardclose(diameter_base->transport->clientsocket);
      diameter_base->transport->clientsocket = -1;
    }

    if (diameter_base->transport->serversocket > 0) {
      hardclose(diameter_base->transport->serversocket);
      diameter_base->transport->serversocket = -1;
    }
  }
  diameter_base->transport->trans_state = TRANSPORT_CLOSED;

  if (AST_PTHREADT_NULL != diameter_base->diameter_thread) {
    pthread_cancel(diameter_base->diameter_thread);
    pthread_kill(diameter_base->diameter_thread, SIGURG);
    pthread_join(diameter_base->diameter_thread, NULL);
    diameter_base->diameter_thread = AST_PTHREADT_NULL;
  }
}

int register_transporttimer(void)
{
  /* Only when the peer has not disconnected try to reconnect with him */
  if (!diameter_base->transport->disconnected) {
    if (-1 == diameter_base->transport_timer_id) {
      /* Restarting(or starting the timer if not present) the timer for attempting to make connections at interval of 30 seconds */
      if (-1 == (diameter_base->transport_timer_id = ast_sched_add(diameter_base->scheduler, 1000*WAIT_TIME, re_connect, NULL))) {
        ast_log(LOG_ERROR, "Failed to start the transport timer.\n");
        return 0;
      }
    }
  }
  return 1;
}

/*! \brief Restarts the transport */
int transport_restart(void)
{
  if (!diameter_cleanup())
    return 0;

  if (diameter_base->conn_initiator) /* When the existing connection is initiator. */
    if (!restart_connecting_socket())
      return 0;

  return process_event(START, NULL);
}


/*! 
 * \brief Listens for and receives the incoming Diameter packets
 *
 * \retval 1  on success and 0 on failure 
 */
static void* transport_tcp_listener(void* pvdata)
{	
	int received = 0, nfds = 0;
	int valopt; 
	int sched_res = 0;
	int connection_close = 0;
	unsigned char buffer[MAX_BUFFER_SIZE] = "\0";
	socklen_t size = sizeof(struct sockaddr);
	fd_set read_fd, except_fd, write_fd;
	struct timeval tv;
	struct sockaddr server_addr;
	socklen_t lon;
	enum transport_state trans_state = TRANSPORT_CLOSED;

	while(1) {

		pthread_testcancel();
		sched_res = ast_sched_wait(diameter_base->scheduler);
		if ((sched_res < 0) || (sched_res > 1000))
			sched_res = 1000;

		tv.tv_sec = 0;
		tv.tv_usec = sched_res;
		FD_ZERO(&read_fd);
	  FD_ZERO(&except_fd);

		if (diameter_base->transport && diameter_base->transport->activesocket > 0) {
			FD_SET(diameter_base->transport->activesocket,&read_fd);
			FD_SET(diameter_base->transport->activesocket,&except_fd);
			if (nfds < diameter_base->transport->activesocket)
				nfds = diameter_base->transport->activesocket;
		} else {
			FD_SET(diameter_base->transport->serversocket,&read_fd);
			FD_SET(diameter_base->transport->serversocket,&except_fd);
			if (nfds < diameter_base->transport->serversocket)
				nfds = diameter_base->transport->serversocket;

			/* If the transport in connecting state then set for write description */
			if (TRANSPORT_CONNECTING ==  diameter_base->transport->trans_state) {
				FD_SET(diameter_base->transport->clientsocket,&write_fd);
				FD_SET(diameter_base->transport->clientsocket,&except_fd);
				if (nfds < diameter_base->transport->clientsocket)
					nfds = diameter_base->transport->clientsocket;
			}
		}
		
		if (select(nfds+1, &read_fd, &write_fd, &except_fd, &tv) > 0) {
			if (diameter_base->transport->trans_state == TRANSPORT_INIT || diameter_base->transport->trans_state == TRANSPORT_CONNECTING) {
				if (FD_ISSET(diameter_base->transport->serversocket, &read_fd)) {
					if (-1 == (diameter_base->transport->activesocket = accept(diameter_base->transport->serversocket, &server_addr, &size))) {
						ast_log(LOG_ERROR, "Accept error is [%s]\n", transport_error_string(errno));	
					} else {
						if (size != sizeof(server_addr) || server_addr.sa_family != AF_INET) {
							diameter_base->transport->activesocket = -1;
    				} else {

							diameter_base->conn_initiator = 0;
							update_transport_state_and_process(TRANSPORT_UP);

							/* Deleting the transport timer when a connection is established */
							if (-1 != diameter_base->transport_timer_id) {
								ast_mutex_lock(&diameter_base->baselock);
								AST_SCHED_DEL(diameter_base->scheduler, diameter_base->transport_timer_id);
								diameter_base->transport_timer_id = -1;
								ast_mutex_unlock(&diameter_base->baselock);
							}
							received = recv(diameter_base->transport->activesocket, buffer, MAX_BUFFER_SIZE-1, 0);
							if (received > 0) {
								buffer[received] = '\0';
								process_diameter_message(buffer, received);
							}
						}
					}
				} else if (FD_ISSET(diameter_base->transport->serversocket, &except_fd)) {
					ast_log(LOG_ERROR, "Exception received in server socket\n");
				} else if (FD_ISSET(diameter_base->transport->clientsocket, &write_fd)) {

					/*Checking for the connect() status for client socket*/
					lon = sizeof(int); 
					getsockopt(diameter_base->transport->clientsocket, SOL_SOCKET, SO_ERROR, (void*)(&valopt), &lon); 

					if (valopt) { 
						ast_log(LOG_ERROR, "getsockopt error is [%s]\n", transport_error_string(errno));	
						/* Connection failed, updating the status to start the reconnect scheduler */
						trans_state = TRANSPORT_RECONNECT;	
					} else {
						diameter_base->transport->activesocket = diameter_base->transport->clientsocket;			
						diameter_base->conn_initiator = 1;
						trans_state = TRANSPORT_UP;
						/* Deleting the transport timer when a connection is established */
						if (-1 != diameter_base->transport_timer_id) {
							ast_mutex_lock(&diameter_base->baselock);
							AST_SCHED_DEL(diameter_base->scheduler, diameter_base->transport_timer_id);
          		ast_mutex_unlock(&diameter_base->baselock);
						}
					}
					update_transport_state_and_process(trans_state);
				} else if (FD_ISSET(diameter_base->transport->clientsocket, &except_fd)) {
					ast_log(LOG_ERROR, "Exception received in server socket\n");
				}
			}	else if (TRANSPORT_UP == diameter_base->transport->trans_state) {
				if (FD_ISSET(diameter_base->transport->activesocket, &read_fd)) {
					received = recv(diameter_base->transport->activesocket, buffer, MAX_BUFFER_SIZE-1, 0);

					if (-1 == received) {
						if (EWOULDBLOCK != errno) {
							connection_close = 1;
						}
    			} else if(0 == received) {   	/* If 0 bytes received means FIN received */
						connection_close = 1;
    			} else if(received > 0) {
						buffer[received] = '\0';
						process_diameter_message(buffer, received );
					}

					if (connection_close) {
						connection_close = 0;
						if (INVALID_SOCKET != diameter_base->transport->activesocket) {
							hardclose(diameter_base->transport->activesocket);
       				diameter_base->transport->activesocket = INVALID_SOCKET;
    				}

						if(0 == update_transport_state_and_process(TRANSPORT_DOWN)) {
   						ast_log(LOG_ERROR, "Failed to restart the transport\n");
						}
					}
				} else if (FD_ISSET(diameter_base->transport->activesocket, &except_fd)) {
					ast_log(LOG_ERROR, "Exception received in packet receiving socket\n");
				}
			}
		}
		ast_mutex_lock(&schedlock);
		sched_res = ast_sched_runq(diameter_base->scheduler);
		ast_mutex_unlock(&schedlock);
	}/* end of while */
	return NULL;
}

/* \brief Makes connection attempts at regular intervals when no connection exists */
static int re_connect(const void *data)
{
	diameter_base->state = CLOSED;
	if (!process_event(START, NULL)) {
		ast_log(LOG_ERROR, "Failed to process event \"START\"\n");
		return -1;
	}

	/* If connection is established then we will remove this timer */
	return !(TRANSPORT_UP == diameter_base->transport->trans_state);
}
	
static int restart_connecting_socket(void) 
{
	int sock_fd = 0;
	
	if (-1 == (sock_fd = create_new_socket()))
		return 0;

	diameter_base->transport->clientsocket = sock_fd;
	return 1;
}

static const char* transport_error_string(int err_no)
{
#define errno_to_str(a, b) case a: return "Socket error " #a ": " b;

  switch(err_no) {
    errno_to_str(EALREADY, "The requested operation was already completed");
    errno_to_str(EINPROGRESS, "A blocking call is in progress");
    errno_to_str(EWOULDBLOCK, "Operation would block");
    errno_to_str(ENOTSOCK, "The descriptor is not a socket");
    errno_to_str(EDESTADDRREQ, "A destination address is required in the function call");
    errno_to_str(EMSGSIZE, "The datagram was too large to fit in a single buffer and was truncated");
    errno_to_str(EPROTOTYPE, "The specified port is the wrong type for this socket");
    errno_to_str(ENOPROTOOPT, "The protocol option is unknown or unsupported");
    errno_to_str(EPROTONOSUPPORT, "The requested protocol is not supported");
    errno_to_str(ESOCKTNOSUPPORT, "The specified socket type is not supported in this address family");
    errno_to_str(EOPNOTSUPP, "The socket option is not supported on this socket");
    errno_to_str(EPFNOSUPPORT, "The requested Protocol Family is not supported for this socket");
    errno_to_str(EAFNOSUPPORT, "The requested Address Family is not supported for this socket");
    errno_to_str(EADDRINUSE, "The requested port/address is already in use");
    errno_to_str(EADDRNOTAVAIL, "The requested address is not available");
    errno_to_str(ENETDOWN, "Network subsystem is down");
    errno_to_str(ENETUNREACH, "The remote network is unreachable");
    errno_to_str(ENETRESET, "WinSock has dropped the connection");
    errno_to_str(ECONNABORTED, "The connection was dropped due to timeout or other failure");
    errno_to_str(ECONNRESET, "The connection has been reset by the remote peer");
    errno_to_str(ENOBUFS, "No buffer space is available");
    errno_to_str(EISCONN, "The specified socket is already connected");
    errno_to_str(ENOTCONN, "The specified socket is not connected");
    errno_to_str(ESHUTDOWN, "The specified socket has already been shut down");
    errno_to_str(ETIMEDOUT, "The connection attempt has timed out");
    errno_to_str(ECONNREFUSED, "The connection attempt was forcefully refused");
    errno_to_str(ENAMETOOLONG, "The specified name was too long");
    errno_to_str(EHOSTDOWN, "The remote host has gone down");
    errno_to_str(EHOSTUNREACH, "The remote host is unreachable");

   default:
     return "Socket error: An unknown socket error has occured";
  }
#undef errno_to_str
}

/* \brief Creates a new socket */
static int create_new_socket(void)
{
  int sock_fd = 0;
  int flags = 0;

  if ((sock_fd = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) < 0 ) {
    ast_log(LOG_ERROR, "Socket create error is [%s]\n", transport_error_string (errno));
    return -1;
  }
  flags = fcntl(sock_fd, F_GETFL);
  fcntl(sock_fd, F_SETFL, flags | O_NONBLOCK);
  return sock_fd;
}

/* \brief Creates sockets for connecting and accepting connection.*/
static int create_endpoint(void)
{
  struct sockaddr_in peer_addr;
  int sock_fd = -1;

  if (-1 == (sock_fd = create_new_socket()))
    return 0;

  diameter_base->transport->clientsocket = sock_fd;

  if (-1 == (sock_fd = create_new_socket()))
    return 0;

  peer_addr.sin_family = AF_INET;
  peer_addr.sin_addr.s_addr = inet_addr(diameter_base->config->host_ip);
  peer_addr.sin_port = htons(DIAMETER_PORT);

  if (-1 == bind(sock_fd, (struct sockaddr *)&peer_addr, sizeof(struct sockaddr))) {
    ast_log(LOG_ERROR, "Bind error is [%s]\n", transport_error_string(errno));
    return 0;
  }

  diameter_base->transport->serversocket = sock_fd;
  return 1;
}

