/*Copyright and Author
 *
 *This program is free software, distributed under the terms of the GNU General Public License Version 2.
 * 
 * \author Leena <leena_m@spanservices.com>, (SPAN Infotech India Pvt. Ltd., Bangalore)
 *
 * \module diameter
 * 
 * \created 24\03\09
 */

/*! \file
  \brief Connection related routines.
  
 */

#ifndef _DIAMETER_TRANSPORT_H
#define _DIAMETER_TRANSPORT_H

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#if defined(__cplusplus) || defined(c_plusplus)
extern "C" {
#endif

#define MAX_BUFFER_SIZE 1024 /*!< Buffer to hold the data */
#define closesocket close

/*!< To set SO_OOBINLINE option for the socket */ 
#ifndef setoobinline
#define setoobinline(socket) { int bBool = 1; setsockopt(socket, SOL_SOCKET, SO_OOBINLINE,(const char* )&bBool,sizeof(bBool)); }
#endif

/*!< To close the socket */
#ifndef hardclose
  #define hardclose(socket) { struct linger Linger; Linger.l_onoff = 1; Linger.l_linger = 0;     setsockopt(socket,SOL_SOCKET,SO_LINGER,(const char* )&Linger,sizeof(Linger)); closesocket(socket); }
#endif


/*! brief Status of the transport */
enum transport_state {
	TRANSPORT_CLOSED,			/*!< No connection exists */
	TRANSPORT_INIT,				/*!< Transport is initialised */
	TRANSPORT_CONNECTING, /*!< When a connection attempt is being made */
	TRANSPORT_RECONNECT,  /*!< When reconnection attempts are made */
	TRANSPORT_UP,         /*!< When the connection is established between the peers */
	TRANSPORT_DOWN				/*!< When a connection is disconnected */
};

struct diameter_peer_transport {
	int clientsocket;									/*!< Socket for making connections */
	int serversocket;									/*!< Socket for accepting the incoming connections */
	int activesocket;									/*!< Socket which actively used for exchanging messages once the connection is up */
	int connected;										/*!< Is high if connection exists else 0 */
	int disconnected;									/*!< Is high when the diameter client is disconnected by its peer */
	enum transport_state trans_state; /*!< Holds the state of the transport */
};

/*! 
 * \brief Creates endpoints for connection and initiates a thread for incoming packets
 *
 * \retval 1 on success and 0 on failure
 */
int diameter_transport_init(void);

/*! 
 * \brief Sends a diameter message
 *
 * \param data the message to be sent
 * \param data_length the length of the message to be sent
 *
 * \retval 1 on success and 0 on failure
 */
int diameter_transport_send(const unsigned char *data, int data_length);

/*!
 * \brief Makes a connection attempt with the diameter peer.
 *
 * \returns the state of the transport 
 */
enum transport_state diameter_transport_connection(void);

/*! 
 * \brief Closes the connection and kills the listener thread(for incoming packets)
 *
 */
void diameter_transport_cleanup(void);

/*!
 * \brief Registers a timer for making connection attempts at regular intervals no connection exists.
 *
 * \retval 1 on success and 0 on failure
 */
int register_transporttimer(void);

/*
 * \brief Closes the existing connection and restarts the transport for a new connection.
 *
 * \retval 1 on success and 0 on failure
 */
int transport_restart(void);

#if defined(__cplusplus) || defined(c_plusplus)
}
#endif


#endif /* _DIAMETER_TRANSPORT_H */

